"""Simplified session management tests."""

import re
import tempfile
import time
from unittest.mock import MagicMock, patch

import pytest

from dotevals.models import Evaluation, EvaluationStatus, Record, Result, Score
from dotevals.sessions import SessionManager, get_git_commit


class TestSessionManager:
    """Core SessionManager tests."""

    def test_init_without_experiment(self, tmp_path):
        """Test initialization without experiment creates ephemeral."""
        import uuid

        eval_name = f"test_eval_{uuid.uuid4().hex[:8]}"
        manager = SessionManager(eval_name, storage=f"json://{tmp_path}")

        assert manager.storage is not None
        assert manager.current_experiment is not None
        # Should create timestamped ephemeral experiment
        assert re.match(r"\d{8}_\d{6}_[a-f0-9]{8}", manager.current_experiment)

    def test_init_with_experiment(self, tmp_path):
        """Test initialization with named experiment."""
        import uuid

        unique_id = uuid.uuid4().hex[:8]
        eval_name = f"test_eval_{unique_id}"
        exp_name = f"test_exp_{unique_id}"
        manager = SessionManager(
            evaluation_name=eval_name,
            experiment_name=exp_name,
            storage=f"json://{tmp_path}",
        )

        assert manager.current_experiment == exp_name
        assert manager.current_evaluation == eval_name
        assert exp_name in manager.storage.list_experiments()

    @patch("dotevals.sessions.get_git_commit")
    def test_start_evaluation(self, mock_git_commit, tmp_path):
        """Test starting an evaluation."""
        mock_git_commit.return_value = "abc123"

        import uuid

        unique_id = uuid.uuid4().hex[:8]
        manager = SessionManager(
            evaluation_name=f"eval1_{unique_id}",
            experiment_name=f"test_exp_{unique_id}",
            storage=f"json://{tmp_path}",
        )

        manager.start_evaluation()

        # Check evaluation was created
        evaluations = manager.storage.list_evaluations(manager.current_experiment)
        assert manager.current_evaluation in evaluations

        # Check evaluation status
        eval_obj = manager.storage.load_evaluation(
            manager.current_experiment, manager.current_evaluation
        )
        assert eval_obj.status == EvaluationStatus.RUNNING
        assert eval_obj.metadata.get("git_commit") == "abc123"

    def test_add_results(self, tmp_path):
        """Test adding results to evaluation."""
        import uuid

        unique_id = uuid.uuid4().hex[:8]
        manager = SessionManager(
            evaluation_name=f"eval1_{unique_id}",
            experiment_name=f"test_exp_{unique_id}",
            storage=f"json://{tmp_path}",
        )

        manager.start_evaluation()

        # Add results
        record = Record(
            item_id=0,
            dataset_row={"input": "test"},
            result=Result(Score("test", True, [])),
        )
        manager.add_results([record])

        # Verify results were stored
        results = manager.storage.get_results(
            manager.current_experiment, manager.current_evaluation
        )
        assert len(results) == 1
        assert results[0].dataset_row["input"] == "test"

    def test_finish_evaluation(self, tmp_path):
        """Test finishing an evaluation."""
        import uuid

        unique_id = uuid.uuid4().hex[:8]
        manager = SessionManager(
            evaluation_name=f"eval1_{unique_id}",
            experiment_name=f"test_exp_{unique_id}",
            storage=f"json://{tmp_path}",
        )

        manager.start_evaluation()
        manager.finish_evaluation()

        # Check status updated
        eval_obj = manager.storage.load_evaluation(
            manager.current_experiment, manager.current_evaluation
        )
        assert eval_obj.status == EvaluationStatus.COMPLETED

    def test_get_results(self, tmp_path):
        """Test retrieving results."""
        import uuid

        unique_id = uuid.uuid4().hex[:8]
        manager = SessionManager(
            evaluation_name=f"eval1_{unique_id}",
            experiment_name=f"test_exp_{unique_id}",
            storage=f"json://{tmp_path}",
        )

        manager.start_evaluation()

        # Add some results
        records = [
            Record(
                item_id=i,
                dataset_row={"idx": i},
                result=Result(Score("test", True, [])),
            )
            for i in range(5)
        ]
        manager.add_results(records)

        # Get results
        retrieved = manager.get_results()
        assert len(retrieved) == 5
        assert all(r.dataset_row["idx"] == i for i, r in enumerate(retrieved))

    def test_evaluation_progress_tracking(self, tmp_path):
        """Test progress tracking during evaluation."""
        import uuid

        unique_id = uuid.uuid4().hex[:8]
        manager = SessionManager(
            evaluation_name=f"eval1_{unique_id}",
            experiment_name=f"test_exp_{unique_id}",
            storage=f"json://{tmp_path}",
        )

        manager.start_evaluation()

        # Track progress manually since evaluation_progress is EvaluationProgress object
        completed = 0

        # Add results incrementally
        for i in range(5):
            record = Record(
                item_id=i,
                dataset_row={"idx": i},
                result=Result(Score("test", True, [])),
            )
            manager.add_results([record])
            completed += 1

        # Check through evaluation_progress object
        assert manager.evaluation_progress.completed_count == 5
        assert manager.evaluation_progress.error_count == 0

    def test_ephemeral_experiment_naming(self, tmp_path):
        """Test ephemeral experiment name generation."""
        import uuid
        from unittest.mock import Mock

        # Generate eval name before patching
        eval_name = f"test_eval_{uuid.uuid4().hex[:8]}"

        with patch("dotevals.sessions.datetime") as mock_datetime:
            mock_datetime.now.return_value.strftime.return_value = "20240101_120000"

            with patch("dotevals.sessions.uuid.uuid4") as mock_uuid4:
                mock_obj = Mock()
                # uuid4() returns object, then we convert to str and slice [:8]
                mock_obj.__str__ = Mock(
                    return_value="abcd1234-efgh-ijkl-mnop-qrstuvwxyz12"
                )
                mock_uuid4.return_value = mock_obj

                manager = SessionManager(eval_name, storage=f"json://{tmp_path}")

                assert manager.current_experiment == "20240101_120000_abcd1234"

    def test_no_storage_url(self, tmp_path):
        """Test initialization without storage URL uses default."""
        import uuid

        eval_name = f"test_eval_{uuid.uuid4().hex[:8]}"
        manager = SessionManager(eval_name, storage=f"json://{tmp_path}")

        # Should create storage
        assert manager.storage is not None

        # Should be able to use it normally
        manager.start_evaluation()
        manager.finish_evaluation()


class TestGitIntegration:
    """Tests for git integration."""

    def test_get_git_commit_success(self):
        """Test getting git commit when in repo."""
        with patch("subprocess.check_output") as mock_check:
            mock_check.return_value = b"abc123def456789\n"

            commit = get_git_commit()

            assert commit == "abc123de"
            mock_check.assert_called_once()

    def test_get_git_commit_not_repo(self):
        """Test getting git commit outside repo."""
        import subprocess

        with patch("subprocess.check_output") as mock_check:
            mock_check.side_effect = subprocess.CalledProcessError(128, "git")

            commit = get_git_commit()

            assert commit is None

    def test_get_git_commit_error(self):
        """Test git commit with subprocess error."""
        with patch("subprocess.check_output") as mock_check:
            mock_check.side_effect = Exception("Git not found")

            # get_git_commit doesn't catch generic exceptions, only CalledProcessError
            # So this will raise
            with pytest.raises(Exception):
                get_git_commit()


@pytest.mark.parametrize(
    "status",
    [
        EvaluationStatus.RUNNING,
        EvaluationStatus.COMPLETED,
        EvaluationStatus.FAILED,
    ],
)
def test_evaluation_status_transitions(tmp_path, status):
    """Test different evaluation status transitions."""
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    manager = SessionManager(
        evaluation_name=f"eval1_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    manager.start_evaluation()

    # Update status
    manager.storage.update_evaluation_status(
        manager.current_experiment, manager.current_evaluation, status
    )

    # Verify
    eval_obj = manager.storage.load_evaluation(
        manager.current_experiment, manager.current_evaluation
    )
    assert eval_obj.status == status


def test_finish_evaluation_with_failure(tmp_path):
    """Test finishing evaluation with failure status."""
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    manager = SessionManager(
        evaluation_name=f"eval1_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    manager.start_evaluation()
    manager.finish_evaluation(success=False)

    # Check status updated to FAILED
    eval_obj = manager.storage.load_evaluation(
        manager.current_experiment, manager.current_evaluation
    )
    assert eval_obj.status == EvaluationStatus.FAILED


def test_add_results_with_errors(tmp_path):
    """Test adding results with errors updates progress tracking."""
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    manager = SessionManager(
        evaluation_name=f"eval1_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    manager.start_evaluation()

    # Add results with errors
    records = [
        Record(
            item_id=0,
            dataset_row={"idx": 0},
            result=Result(Score("test", True, [])),
            error="Error 1",
        ),
        Record(
            item_id=1, dataset_row={"idx": 1}, result=Result(Score("test", True, []))
        ),
        Record(
            item_id=2,
            dataset_row={"idx": 2},
            result=Result(Score("test", True, [])),
            error="Error 2",
        ),
    ]

    manager.add_results(records)

    # Check error tracking
    assert manager.evaluation_progress.completed_count == 3
    assert manager.evaluation_progress.error_count == 2


def test_session_without_start_evaluation(tmp_path):
    """Test adding results without calling start_evaluation."""
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    exp_name = f"test_exp_{unique_id}"
    eval_name = f"eval1_{unique_id}"
    manager = SessionManager(
        evaluation_name=eval_name,
        experiment_name=exp_name,
        storage=f"json://{tmp_path}",
    )

    # Don't call start_evaluation - evaluation_progress is None
    assert manager.evaluation_progress is None

    # Create evaluation manually so storage doesn't complain
    eval = Evaluation(
        evaluation_name=eval_name,
        started_at=time.time(),
        status=EvaluationStatus.RUNNING,
    )
    manager.storage.create_evaluation(exp_name, eval)

    # Should handle gracefully without evaluation_progress
    record = Record(
        item_id=0,
        dataset_row={"test": "data"},
        result=Result(Score("test", True, [])),
        error="Error",
    )

    # This should not crash even though evaluation_progress is None
    manager.add_results([record])

    # Results should still be stored
    results = manager.storage.get_results(exp_name, eval_name)
    assert len(results) == 1


def test_resume_evaluation_with_existing_results(tmp_path):
    """Test resuming an evaluation that already has results."""
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    exp_name = f"test_exp_{unique_id}"
    eval_name = f"eval1_{unique_id}"

    # First session
    manager1 = SessionManager(
        evaluation_name=eval_name,
        experiment_name=exp_name,
        storage=f"json://{tmp_path}",
    )
    manager1.start_evaluation()

    records = [
        Record(
            item_id=i, dataset_row={"idx": i}, result=Result(Score("test", True, []))
        )
        for i in range(5)
    ]
    manager1.add_results(records)
    # Don't finish - simulate crash

    # Second session resumes
    manager2 = SessionManager(
        evaluation_name=eval_name,
        experiment_name=exp_name,
        storage=f"json://{tmp_path}",
    )

    # Should detect existing evaluation (prints message)
    import io
    import sys

    captured = io.StringIO()
    sys.stdout = captured
    manager2.start_evaluation()
    sys.stdout = sys.__stdout__

    output = captured.getvalue()
    assert "Resuming from 5 completed samples" in output

    # Add more results
    more_records = [
        Record(
            item_id=i, dataset_row={"idx": i}, result=Result(Score("test", True, []))
        )
        for i in range(5, 10)
    ]
    manager2.add_results(more_records)

    # All results present
    final_results = manager2.get_results()
    assert len(final_results) == 10


def test_concurrent_result_addition(tmp_path):
    """Test adding results concurrently."""
    import threading
    import uuid

    unique_id = uuid.uuid4().hex[:8]
    manager = SessionManager(
        evaluation_name=f"eval1_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    manager.start_evaluation()

    def add_batch(start_idx):
        """Add a batch of results."""
        records = [
            Record(
                item_id=start_idx + i,
                dataset_row={"idx": start_idx + i},
                result=Result(Score("test", True, [])),
            )
            for i in range(10)
        ]
        manager.add_results(records)

    # Start multiple threads
    threads = []
    for i in range(0, 50, 10):
        t = threading.Thread(target=add_batch, args=(i,))
        threads.append(t)
        t.start()

    # Wait for completion
    for t in threads:
        t.join()

    # Verify all results
    results = manager.get_results()
    assert len(results) == 50
    item_ids = [r.item_id for r in results]
    assert sorted(item_ids) == list(range(50))



"""Execution runners for dotevals evaluations."""

import asyncio
from typing import Any, Optional

import pytest

from dotevals.progress import MultiProgress


class Runner:
    """Base runner for dotevals evaluations.

    The Runner class handles all evaluation orchestration including:
    - Sequential and concurrent execution strategies
    - Fixture lifecycle management
    - Progress tracking and reporting
    - Result collection

    For resource management (model clients, database connections, etc.), use
    ModelProvider plugins distributed as pytest fixtures.

    Examples:
        Direct usage:
        ```python
        runner = Runner(
            experiment_name="my_experiment",
            samples=100,
            concurrent=True
        )
        await runner.run_evaluations(evaluation_items)
        ```

        The Runner handles all orchestration logic internally, so subclasses
        typically don't need to override any methods.
    """

    def __init__(
        self,
        experiment_name: str | None = None,
        samples: int | None = None,
        storage: str | None = None,
        concurrent: bool = True,
        results_dict: dict[str, Any] | None = None,
    ) -> None:
        """Initialize runner with evaluation parameters.

        Args:
            experiment_name: Name of the experiment
            samples: Number of samples to evaluate
            storage: Storage backend path (e.g., 'json://.dotevals', 'sqlite://results.db')
            concurrent: Whether to run async evaluations concurrently
            results_dict: Optional dictionary to store evaluation results
        """
        self.experiment_name = experiment_name
        self.samples = samples
        self.storage = storage
        self.concurrent = concurrent
        self.results = results_dict if results_dict is not None else {}
        self.progress_manager = MultiProgress()

    async def run_evaluations(self, evaluation_items: list[pytest.Item]) -> None:
        """Run all evaluation items with common orchestration logic.

        This method provides the standard orchestration for running evaluations,
        handling both sequential and concurrent execution.

        Args:
            evaluation_items: List of pytest items representing evaluations
        """
        if not evaluation_items:
            return

        # Separate items by execution strategy
        sequential_items = []
        concurrent_items = []

        for item in evaluation_items:
            if self.concurrent and _is_async_evaluation(item):
                concurrent_items.append(item)
            else:
                sequential_items.append(item)

        # Start progress display
        total_items = len(sequential_items) + len(concurrent_items)
        self.progress_manager.start(total_items)

        try:
            # Run sequential items first
            for item in sequential_items:
                await self._run_single_evaluation(item)

            # Run concurrent items together
            if concurrent_items:
                tasks = []
                for item in concurrent_items:
                    task = asyncio.create_task(self._run_single_evaluation(item))
                    tasks.append(task)
                await asyncio.gather(*tasks)
        finally:
            self.progress_manager.finish()

    async def _run_single_evaluation(self, item: pytest.Item) -> None:
        """Execute a single evaluation item with fixture management.

        This method handles the execution of individual evaluations, including
        creating fixtures before and tearing them down after the evaluation,
        properly handling both sync and async functions.

        Args:
            item: The pytest Item representing a single evaluation
        """
        from dotevals.fixtures import FixtureManager
        from dotevals.sessions import SessionManager

        eval_fn = item.function

        # Create SessionManager for this evaluation
        session_manager = SessionManager(
            evaluation_name=item.name,
            experiment_name=self.experiment_name,
            storage=self.storage,
        )

        # Create fixtures for this evaluation. They will be torn down
        # after the evaluation has run.
        fixture_kwargs, request = await FixtureManager.create_fixtures(item)

        try:
            if asyncio.iscoroutinefunction(eval_fn):
                result = await eval_fn(
                    session_manager=session_manager,
                    samples=self.samples,
                    progress_manager=self.progress_manager,
                    **fixture_kwargs,
                )
            else:
                result = eval_fn(
                    session_manager=session_manager,
                    samples=self.samples,
                    progress_manager=self.progress_manager,
                    **fixture_kwargs,
                )
                if asyncio.iscoroutine(result):
                    result = await result

            self.results[item.name] = result
        finally:
            await FixtureManager.teardown_fixtures(request)


def _is_async_evaluation(item: pytest.Item) -> bool:
    """Check if an evaluation item is async."""
    eval_fn = item.function
    original_func = getattr(eval_fn, "__wrapped__", eval_fn)
    return asyncio.iscoroutinefunction(original_func)



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from dotevals.core import DatasetInfo
    from dotevals.models import Record

from rich.console import Console
from rich.live import Live
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Table
from rich.text import Text


def _is_running_under_pytest() -> bool:
    """Check if we're currently running under pytest.

    Returns:
        bool: True if running under pytest, False otherwise.

    Note:
        This is used to disable progress bars during pytest execution
        to avoid interfering with test output.
    """
    try:
        import sys

        return "pytest" in sys.modules
    except ImportError:
        return False


class ProgressTracker:
    """Self-contained progress tracker for a single evaluation.

    This class encapsulates all logic for tracking a single evaluation's progress,
    including completion counting, error tracking, and metrics calculation.
    It is responsible for its own state management and formatting.

    Attributes:
        evaluation_name (str): Name of the evaluation being tracked
        total_items (int, optional): Total number of items to process
        completed (int): Number of completed items
        error_count (int): Number of items that resulted in errors
        task_id (str, optional): Rich progress task ID for multi-evaluation mode
        metrics (dict): Live metrics for single evaluation mode
    """

    def __init__(
        self,
        evaluation_name: str,
        total_items: int | None = None,
        task_id: str | None = None,
    ) -> None:
        """Initialize progress tracker.

        Args:
            evaluation_name (str): Name of the evaluation being tracked
            total_items (int, optional): Total number of items expected.
                If None, progress is tracked as a spinner.
            task_id (str, optional): Rich progress task ID for multi-evaluation mode
        """
        self.evaluation_name = evaluation_name
        self.total_items = total_items
        self.completed = 0
        self.error_count = 0
        self.task_id = task_id
        self.metrics: dict[str, float] = {}
        self._all_results: list[Record] = []  # For metrics calculation

    def update_progress(
        self, result: Optional["Record"] = None, completed_count: int | None = None
    ) -> None:
        """Update progress for this evaluation.

        Args:
            result (Record, optional): Latest evaluation result
            completed_count (int, optional): Set absolute completion count
        """
        if completed_count is not None:
            self.completed = completed_count
        else:
            self.completed += 1

        # Track errors
        if result and hasattr(result, "error") and result.error is not None:
            self.error_count += 1

        # Update metrics
        if result:
            self._update_metrics(result)

    def _update_metrics(self, result: "Record") -> None:
        """Calculate live metrics for this evaluation.

        Args:
            result (Record): The latest evaluation result to include
        """
        self._all_results.append(result)

        # Determine expected score structure from successful results
        expected_scores = []
        for res in self._all_results:
            if (
                hasattr(res, "error")
                and res.error is None
                and hasattr(res, "result")
                and res.result.scores
            ):
                expected_scores = [
                    (score.name, score.metrics)
                    for score in res.result.scores
                    if hasattr(score, "name")
                ]
                break

        # Recalculate metrics from all results
        aggregated_results: dict[str, dict[Any, list[Any]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for res in self._all_results:
            if hasattr(res, "error") and res.error is not None:
                # For error cases, add False values for each expected score
                for score_name, metrics in expected_scores:
                    for metric in metrics:
                        aggregated_results[score_name][metric].append(False)
            else:
                # For successful cases, use actual scores
                if hasattr(res, "result") and res.result.scores:
                    for score in res.result.scores:
                        for metric in score.metrics:
                            aggregated_results[score.name][metric].append(score.value)

        # Compute and store metrics
        computed_metrics = {}
        for evaluator_name, metrics_values in aggregated_results.items():
            for metric_func, values in metrics_values.items():
                if values:  # Only compute if we have values
                    metric_name = metric_func.__name__.replace("_", " ").title()
                    computed_metrics[metric_name] = metric_func(values)

        self.metrics = computed_metrics

    def get_error_text(self) -> str:
        """Get formatted error text for display.

        Returns:
            str: Formatted error text or empty string if no errors
        """
        if self.error_count > 0:
            return f" ! {self.error_count} error{'s' if self.error_count != 1 else ''}"
        return ""

    def format_metrics_for_display(self) -> str:
        """Format metrics for display in single evaluation mode.

        Returns:
            str: Formatted metrics text or empty string if no metrics
        """
        if not self.metrics:
            return ""

        metrics_parts = []
        for name, value in self.metrics.items():
            if isinstance(value, float):
                metrics_parts.append(f"{name}: {value:.1%}")
            else:
                metrics_parts.append(f"{name}: {value}")

        return " • ".join(metrics_parts) if metrics_parts else ""


class BaseProgressManager(ABC):
    """Base class defining the progress manager interface.

    This defines the common contract that all progress managers must implement,
    while allowing different implementations for different use cases.
    """

    @abstractmethod
    def start_evaluation(
        self,
        evaluation_name: str,
        total_items: int | None = None,
        dataset_info: Any = None,
    ) -> None:
        """Start tracking progress for a specific evaluation."""
        pass

    @abstractmethod
    def update_evaluation_progress(
        self,
        evaluation_name: str,
        completed_count: int | None = None,
        result: Optional["Record"] = None,
    ) -> None:
        """Update progress for a specific evaluation."""
        pass

    @abstractmethod
    def finish(self) -> None:
        """Finish progress tracking and clean up resources."""
        pass

    def __enter__(self) -> "BaseProgressManager":
        """Enter context manager."""
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit context manager."""
        pass


class MultiProgress(BaseProgressManager):
    """Progress display for multiple evaluations (pytest mode).

    This manager handles multiple concurrent evaluations, displaying
    each as a separate progress bar in a shared Rich Progress display.
    Designed for pytest execution where multiple evaluation functions
    run in sequence or parallel.

    Attributes:
        evaluations (dict): Map of evaluation names to ProgressTracker instances
        active (bool): Whether progress tracking is currently active
        progress (Progress): Rich Progress instance for display
        live (Live): Rich Live instance for real-time updates
    """

    def __init__(self):
        """Initialize multi-evaluation progress manager."""
        self.evaluations = {}
        self.active = False
        self.progress = None
        self.live = None

    def start(self, total_evaluations=None):
        """Start the multi-evaluation progress display.

        Args:
            total_evaluations (int, optional): Expected number of evaluations.
        """
        self.active = True
        self.total_evaluations = total_evaluations

        # Set up multi-evaluation display
        self.progress = Progress(
            TextColumn("✨ {task.description}"),
            GreenCompletedBarColumn(bar_width=60),
            MofNCompleteColumn(),
            TextColumn("• Elapsed:"),
            TimeElapsedColumn(),
            TextColumn("• ETA:"),
            TimeRemainingColumn(),
            TextColumn("[bold red]{task.fields[error_text]}[/bold red]"),
        )
        self.live = Live(self.progress, console=Console(), refresh_per_second=10)
        self.live.start()

    def start_evaluation(self, evaluation_name, total_items=None, dataset_info=None):
        """Add a new evaluation to the multi-evaluation display.

        Args:
            evaluation_name (str): Unique name for this evaluation
            total_items (int, optional): Total number of items to process.
            dataset_info (dict, optional): Additional dataset information (unused).
        """
        task_id = None
        if self.progress:
            task_id = self.progress.add_task(
                evaluation_name,
                total=total_items if total_items else 100,
                error_text="",
            )

        tracker = ProgressTracker(evaluation_name, total_items, task_id)
        self.evaluations[evaluation_name] = tracker

    def update_evaluation_progress(
        self, evaluation_name, completed_count=None, result=None
    ):
        """Update progress for a specific evaluation.

        Args:
            evaluation_name (str): Name of the evaluation to update
            completed_count (int, optional): Set absolute completion count.
            result (Record, optional): The evaluation result object.
        """
        if evaluation_name not in self.evaluations:
            return

        tracker = self.evaluations[evaluation_name]
        tracker.update_progress(result=result, completed_count=completed_count)

        # Update the Rich progress display
        if self.progress and tracker.task_id is not None:
            self.progress.update(
                tracker.task_id,
                completed=tracker.completed,
                error_text=tracker.get_error_text(),
            )

    def finish(self):
        """Finish progress tracking and clean up resources."""
        self.active = False
        self.evaluations.clear()

        # Stop display
        if self.live:
            self.live.stop()
            self.live = None
        self.progress = None


class SingleProgress(BaseProgressManager):
    """Progress display for single evaluation with metrics (programmatic mode).

    This manager handles a single evaluation, displaying a progress bar
    with live metrics underneath. Designed for programmatic execution
    where one evaluation runs at a time with detailed feedback.

    Supports context manager usage for automatic cleanup.

    Attributes:
        tracker (ProgressTracker): The single evaluation tracker
        progress (Progress): Rich Progress instance for display
        live (Live): Rich Live instance for real-time updates
    """

    def __init__(self):
        """Initialize single evaluation progress manager."""
        self.tracker = None
        self.progress = None
        self.live = None

    def start_evaluation(self, evaluation_name, total_items=None, dataset_info=None):
        """Start tracking progress for the single evaluation.

        Args:
            evaluation_name (str): Name for this evaluation
            total_items (int, optional): Total number of items to process.
            dataset_info (dict, optional): Additional dataset information for display.
        """
        self.tracker = ProgressTracker(evaluation_name, total_items)

        # Don't show progress bars when running under pytest
        if not _is_running_under_pytest():
            self._setup_display(evaluation_name, total_items, dataset_info)

    def _setup_display(self, evaluation_name, total_items, dataset_info):
        """Set up the Rich progress display.

        Args:
            evaluation_name (str): Name of the evaluation
            total_items (int, optional): Total number of items
            dataset_info (dict, optional): Dataset information for display
        """
        total_count = total_items
        if not total_count and dataset_info:
            total_count = dataset_info.get("total_rows")

        if total_count:
            # Known size - use progress bar with ETA
            self.progress = Progress(
                TextColumn("✨ {task.description}"),
                GreenCompletedBarColumn(bar_width=60),
                MofNCompleteColumn(),
                TextColumn("• Elapsed:"),
                TimeElapsedColumn(),
                TextColumn("• ETA:"),
                TimeRemainingColumn(),
                TextColumn("[bold red]{task.fields[error_text]}[/bold red]"),
                console=Console(),
                expand=False,
            )
            self.main_task = self.progress.add_task(
                evaluation_name, total=total_count, error_text=""
            )
        else:
            # Unknown size - use spinner with completion counter
            self.progress = Progress(
                SpinnerColumn(),
                TextColumn("✨ {task.description}"),
                TextColumn("• {task.completed:,} completed"),
                TimeElapsedColumn(),
                TextColumn("[bold red]{task.fields[error_text]}[/bold red]"),
                console=Console(),
                expand=False,
            )
            self.main_task = self.progress.add_task(
                evaluation_name, total=None, error_text=""
            )

        self.live = Live(self._build_display(), console=Console(), refresh_per_second=4)
        self.live.start()

    def update_evaluation_progress(
        self, evaluation_name, completed_count=None, result=None
    ):
        """Update progress for the single evaluation.

        Args:
            evaluation_name (str): Name of the evaluation (must match tracker)
            completed_count (int, optional): Set absolute completion count.
            result (Record, optional): The evaluation result object.
        """
        if not self.tracker or self.tracker.evaluation_name != evaluation_name:
            return

        self.tracker.update_progress(result=result, completed_count=completed_count)

        # Update the Rich progress display
        if self.progress and hasattr(self, "main_task"):
            self.progress.update(
                self.main_task,
                completed=self.tracker.completed,
                error_text=self.tracker.get_error_text(),
            )

            if self.live:
                self.live.update(self._build_display())

    def _build_display(self):
        """Build the combined progress and metrics display.

        Returns:
            Table: Rich table containing progress bar and metrics
        """
        if not self.progress:
            return ""

        table = Table.grid(padding=(0, 0))
        table.add_column()

        # Add main progress
        table.add_row(self.progress)

        # Add metrics row if available
        if self.tracker:
            metrics_text = self.tracker.format_metrics_for_display()
            if metrics_text:
                table.add_row(Text(metrics_text, style="dim"))

        return table

    def finish(self):
        """Finish progress tracking and clean up resources."""
        # Stop display
        if self.live:
            self.live.stop()
            self.live = None
        self.progress = None
        self.tracker = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager and clean up resources."""
        self.finish()


class GreenCompletedBarColumn(BarColumn):
    """A progress bar column that stays green when completed.

    This custom Rich progress bar column ensures that completed progress
    bars remain green instead of reverting to the default style.
    """

    def __init__(self, bar_width=60):
        """Initialize green completed bar column.

        Args:
            bar_width (int): Width of the progress bar in characters (default: 60)
        """
        super().__init__(
            bar_width=bar_width, complete_style="green", finished_style="green"
        )


def get_dataset_info(dataset: Any) -> "DatasetInfo":
    """Extract dataset information for progress display.

    Args:
        dataset: Dataset object, iterator, or any iterable

    Returns:
        dict: Dictionary containing dataset information with keys:
            - 'name' (str): Display name for the dataset
            - 'total_rows' (int or None): Total number of rows if determinable
    """
    info: DatasetInfo = {"name": "Dataset", "total_rows": None}

    # Check if this is a Dataset instance from the registry
    if hasattr(dataset, "name") and hasattr(dataset, "num_rows"):
        info.update(
            {
                "name": dataset.name.upper(),
                "total_rows": dataset.num_rows,
            }
        )
    # Fallback to size estimation for direct iterators
    elif hasattr(dataset, "__len__"):
        try:
            info["total_rows"] = len(dataset)
        except (TypeError, AttributeError):
            pass

    return info



from .base import Storage, registry

# Expose commonly used functions for convenience
get_storage = registry.get_storage
list_backends = registry.list_backends
register = registry.register

__all__ = ["Storage", "registry", "get_storage", "list_backends", "register"]



"""Adaptive concurrency strategy for async functions."""

import asyncio
import time
from collections import deque
from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from typing import (
    Optional,
    TypeVar,
)

T = TypeVar("T")


class ThroughputTracker:
    """Tracks throughput metrics for the adaptive strategy."""

    def __init__(self, window_size: int = 20):
        self.window_size = window_size
        self.completion_times: deque[float] = deque(maxlen=window_size)
        self.window_start_time: float | None = None
        self.total_completed = 0

    def record_completion(self, timestamp: float) -> None:
        """Record a task completion."""
        self.completion_times.append(timestamp)
        self.total_completed += 1
        if self.window_start_time is None:
            self.window_start_time = timestamp

    def get_throughput(self) -> float | None:
        """Calculate current throughput in requests per second."""
        if len(self.completion_times) < 2:
            return None

        # Calculate throughput over the time window
        time_span = self.completion_times[-1] - self.completion_times[0]
        if time_span <= 0:
            return None

        return (len(self.completion_times) - 1) / time_span

    def get_recent_throughput(self, last_n: int = 5) -> float | None:
        """Get throughput for the last N completions."""
        if len(self.completion_times) < min(last_n, 2):
            return None

        recent = list(self.completion_times)[-last_n:]
        time_span = recent[-1] - recent[0]
        if time_span <= 0:
            return None

        return (len(recent) - 1) / time_span


class Adaptive:
    """Adaptive concurrency strategy that dynamically adjusts to maximize throughput.

    This strategy starts with a conservative concurrency level and continuously
    adjusts it based on observed throughput. It uses a hill-climbing approach
    to find the optimal concurrency level for the given workload and system.
    """

    def __init__(
        self,
        initial_concurrency: int = 5,
        min_concurrency: int = 1,
        max_concurrency: int = 100,
        adaptation_interval: float = 2.0,
        throughput_window: int = 20,
        increase_threshold: float = 0.98,
        decrease_threshold: float = 0.90,
        stability_window: int = 3,
        error_backoff_factor: float = 0.7,
    ):
        """Initialize the adaptive strategy.

        Args:
            initial_concurrency: Starting concurrency level
            min_concurrency: Minimum allowed concurrency
            max_concurrency: Maximum allowed concurrency
            adaptation_interval: Seconds between adaptation decisions
            throughput_window: Number of completions to track for throughput
            increase_threshold: Increase if current/previous throughput > this
            decrease_threshold: Decrease if current/previous throughput < this
            stability_window: Number of measurements before changing direction
            error_backoff_factor: Multiply concurrency by this on errors
        """
        self.current_concurrency = initial_concurrency
        self.min_concurrency = min_concurrency
        self.max_concurrency = max_concurrency
        self.adaptation_interval = adaptation_interval
        self.increase_threshold = increase_threshold
        self.decrease_threshold = decrease_threshold
        self.stability_window = stability_window
        self.error_backoff_factor = error_backoff_factor

        # Tracking state
        self.throughput_tracker = ThroughputTracker(throughput_window)
        self.last_adaptation_time = time.time()
        self.last_throughput: float | None = None
        self.consecutive_increases = 0
        self.consecutive_decreases = 0
        self.recent_errors = 0
        self.total_tasks = 0
        self.adaptation_history: list[tuple[float, int, float]] = []

    async def execute(
        self,
        tasks: Iterator[Callable[[], Awaitable[T]]],
        progress_callback: Callable[[T], None] | None = None,
    ) -> AsyncIterator[T]:
        """Execute tasks with adaptive concurrency control.

        Args:
            tasks: An iterator of callables that return awaitable results
            progress_callback: Optional callback to report progress

        Yields:
            Results from executing the tasks
        """
        tasks_iter = iter(tasks)
        pending_tasks = set()
        adaptation_task = asyncio.create_task(self._adaptation_loop())

        try:
            # Initial fill
            for _ in range(self.current_concurrency):
                try:
                    task_func = next(tasks_iter)
                    task: asyncio.Task = asyncio.create_task(
                        self._execute_with_tracking(task_func)
                    )
                    pending_tasks.add(task)
                    self.total_tasks += 1
                except StopIteration:
                    break

            # Main execution loop
            while pending_tasks:
                done, pending_tasks = await asyncio.wait(
                    pending_tasks, return_when=asyncio.FIRST_COMPLETED
                )

                for completed_task in done:
                    try:
                        result, error = await completed_task
                        if error:
                            self.recent_errors += 1
                        else:
                            self.throughput_tracker.record_completion(time.time())

                        if progress_callback and result is not None:
                            progress_callback(result)

                        if result is not None:
                            yield result
                    except Exception:
                        self.recent_errors += 1
                        raise

                # Add new tasks up to current concurrency
                current_count = len(pending_tasks)
                tasks_to_add = self.current_concurrency - current_count

                for _ in range(tasks_to_add):
                    try:
                        task_func = next(tasks_iter)
                        task = asyncio.create_task(
                            self._execute_with_tracking(task_func)
                        )
                        pending_tasks.add(task)
                        self.total_tasks += 1
                    except StopIteration:
                        break

        finally:
            adaptation_task.cancel()
            for task in pending_tasks:
                task.cancel()
            try:
                await adaptation_task
            except asyncio.CancelledError:
                pass

    async def _execute_with_tracking(
        self, task_func: Callable[[], Awaitable[T]]
    ) -> tuple[T | None, bool]:
        """Execute a task and track its completion."""
        try:
            result = await task_func()
            return result, False
        except Exception:
            # Re-raise the exception so it can be handled in the main loop
            raise

    async def _adaptation_loop(self) -> None:
        """Background task that periodically adjusts concurrency."""
        while True:
            await asyncio.sleep(self.adaptation_interval)
            self._adapt_concurrency()

    def _adapt_concurrency(self) -> None:
        """Adjust concurrency based on throughput measurements."""
        current_throughput = self.throughput_tracker.get_throughput()
        if current_throughput is None:
            return

        # Record adaptation history
        self.adaptation_history.append(
            (time.time(), self.current_concurrency, current_throughput)
        )

        # Handle error backoff
        if self.recent_errors > 0:
            self._decrease_concurrency(
                factor=self.error_backoff_factor**self.recent_errors
            )
            self.recent_errors = 0  # Reset error count
            return

        # Compare with previous throughput
        if self.last_throughput is not None:
            ratio = current_throughput / self.last_throughput

            if ratio > self.increase_threshold:
                # Throughput improved, try increasing
                self.consecutive_decreases = 0
                self.consecutive_increases += 1

                if self.consecutive_increases >= self.stability_window:
                    self._increase_concurrency()
                    self.consecutive_increases = 0

            elif ratio < self.decrease_threshold:
                # Throughput degraded, decrease
                self.consecutive_increases = 0
                self.consecutive_decreases += 1

                if self.consecutive_decreases >= self.stability_window:
                    self._decrease_concurrency()
                    self.consecutive_decreases = 0

            else:
                # Throughput is stable, reset counters
                self.consecutive_increases = 0
                self.consecutive_decreases = 0

        self.last_throughput = current_throughput

    def _increase_concurrency(self) -> None:
        """Increase concurrency level."""
        # Use multiplicative increase with a cap
        if self.current_concurrency < 10:
            self.current_concurrency += 2
        elif self.current_concurrency < 50:
            self.current_concurrency = int(self.current_concurrency * 1.2)
        else:
            self.current_concurrency = int(self.current_concurrency * 1.1)

        self.current_concurrency = min(self.current_concurrency, self.max_concurrency)

    def _decrease_concurrency(self, factor: float = 0.8) -> None:
        """Decrease concurrency level."""
        self.current_concurrency = int(self.current_concurrency * factor)
        self.current_concurrency = max(self.current_concurrency, self.min_concurrency)

    def get_stats(self) -> dict:
        """Get current adaptation statistics."""
        return {
            "current_concurrency": self.current_concurrency,
            "throughput": self.throughput_tracker.get_throughput(),
            "total_completed": self.throughput_tracker.total_completed,
            "total_tasks": self.total_tasks,
            "recent_errors": self.recent_errors,
            "adaptation_history": self.adaptation_history[-10:],  # Last 10 adaptations
        }



"""Batch execution strategy for sync functions."""

from collections.abc import Callable, Iterator
from typing import TypeVar

T = TypeVar("T")


class Batch:
    """Batch execution strategy for sync functions.

    Groups tasks into batches and executes them sequentially within each batch.
    This is useful for processing large datasets in manageable chunks.
    """

    def __init__(self, batch_size: int = 100):
        """Initialize the batch strategy.

        Args:
            batch_size: Number of tasks to execute in each batch
        """
        self.batch_size = batch_size

    def execute(
        self,
        tasks: Iterator[Callable[[], T]],
        progress_callback: Callable[[T], None] | None = None,
    ) -> Iterator[T]:
        """Execute sync tasks in batches.

        Args:
            tasks: An iterator of callables that return results
            progress_callback: Optional callback to report progress

        Yields:
            Results from executing the tasks
        """
        batch = []
        for task in tasks:
            batch.append(task)
            if len(batch) >= self.batch_size:
                # Execute the batch
                for task_func in batch:
                    result = task_func()
                    if progress_callback:
                        progress_callback(result)
                    yield result
                batch = []

        # Execute remaining tasks in the last batch
        for task_func in batch:
            result = task_func()
            if progress_callback:
                progress_callback(result)
            yield result



"""Model providers for dotevals - manage model clients, connections, etc.

Model providers are distributed as pytest fixtures and discovered via entry points.
They handle resource lifecycle (creation, caching, teardown) independently of
execution environment (runners).
"""

import importlib.metadata
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

import pytest


@dataclass
class ModelHandle:
    """Handle to a model with its client and lifecycle methods.

    All model providers MUST return this from their setup() method.
    """

    resource_id: str
    model: Any
    manager: "ModelProvider"
    _active: bool = True

    async def teardown(self) -> None:
        """Tear down this specific model resource."""
        if self._active:
            await self.manager.teardown(self.resource_id)
            self._active = False

    @property
    def is_active(self) -> bool:
        """Check if the model resource is still active."""
        return self._active


class ModelProvider(ABC):
    """Base class for all model providers.

    All model providers MUST inherit from this class and implement
    the required methods.

    The standard pattern is:
        ```python
        handle = await provider.setup(model_spec)
        # use handle.model
        await handle.teardown()
        ```
    """

    @abstractmethod
    async def setup(self, resource_spec: str, **config: Any) -> ModelHandle:
        """Set up a model and return a handle.

        Args:
            resource_spec: Specification for the model (e.g., model name)
            **config: Additional configuration parameters

        Returns:
            ModelHandle containing the model client and teardown method

        Example:
            handle = await manager.setup("gpt-4", temperature=0.7)
            response = await handle.model.generate("Hello")
            await handle.teardown()
        """
        pass

    @abstractmethod
    async def teardown(self, resource_id: str) -> None:
        """Tear down a specific resource.

        Args:
            resource_id: The ID of the resource to tear down

        Note:
            This is typically called by the ModelHandle.teardown() method.
        """
        pass


def discover_model_providers() -> dict[str, str]:
    """Discover all installed model provider plugins.

    Model providers are registered via entry points in the "dotevals.model_providers"
    group. Each entry point should map to a fixture name, the individual provider
    fixture defined in the plugin.

    Returns:
        Dictionary mapping provider name to fixture name

    Example:
        In pyproject.toml:
        ```toml
        [project.entry-points."dotevals.model_providers"]
        vllm = "doteval_vllm:vllm_provider"
        ```
    """
    providers = {}

    try:
        # Try the new API first (Python 3.10+)
        entry_points = importlib.metadata.entry_points(group="dotevals.model_providers")
    except TypeError:
        # Fall back to older API
        all_entry_points = importlib.metadata.entry_points()
        if isinstance(all_entry_points, dict):
            # Old API returns dict mapping group names to lists of entry points
            entry_points = all_entry_points.get("dotevals.model_providers", [])  # type: ignore[arg-type]
        else:
            entry_points = []

    for entry_point in entry_points:
        try:
            # The value should be a fixture name
            fixture_name = entry_point.value.split(":")[-1]
            providers[entry_point.name] = fixture_name
        except Exception as e:
            warnings.warn(f"Failed to process model provider {entry_point.name}: {e}")

    return providers


@pytest.fixture(scope="session")
def model_providers(request: pytest.FixtureRequest) -> dict[str, Any]:
    """Auto-discover and provide access to all model providers.

    This fixture discovers all installed model provider plugins and
    provides access to them via a dictionary.

    Returns:
        Dictionary mapping provider name to provider instance

    Example:
        ```python
        def test_with_providers(model_providers):
            vllm_provider = model_providers["vllm"]
            handle = vllm_provider.setup("llama-7b")
            yield handle.model()
            handle.teardown()
        ```
    """
    discovered = discover_model_providers()
    providers = {}

    for name, fixture_name in discovered.items():
        try:
            # Get the fixture value
            if hasattr(request, "getfixturevalue"):
                provider = request.getfixturevalue(fixture_name)
                providers[name] = provider
        except Exception as e:
            warnings.warn(f"Failed to load model provider {name}: {e}")

    return providers



"""Sliding window concurrency strategy for async functions."""

import asyncio
from collections.abc import AsyncIterator, Callable, Iterator
from typing import TypeVar

T = TypeVar("T")


class SlidingWindow:
    """Sliding window implementation for async concurrency."""

    def __init__(self, max_concurrency: int = 10):
        """Initialize the sliding window strategy.

        Args:
            max_concurrency: Maximum number of concurrent tasks
        """
        self.max_concurrency = max_concurrency

    async def execute(
        self,
        tasks: Iterator[Callable[[], T]],
        progress_callback: Callable[[T], None] | None = None,
    ) -> AsyncIterator[T]:
        """Execute tasks with sliding window concurrency control.

        This maintains a sliding window of concurrent tasks, keeping up to
        max_concurrency tasks running at all times.

        Args:
            tasks: An iterator of callables that return awaitable results
            progress_callback: Optional callback to report progress

        Yields:
            Results from executing the tasks
        """
        tasks_iter = iter(tasks)
        pending_tasks = set()

        try:
            # Fill the initial window to max_concurrency
            for _ in range(self.max_concurrency):
                try:
                    task_func = next(tasks_iter)
                    task: asyncio.Task = asyncio.create_task(task_func())  # type: ignore
                    pending_tasks.add(task)
                except StopIteration:
                    break

            # Process tasks as they complete
            while pending_tasks:
                done, pending_tasks = await asyncio.wait(
                    pending_tasks, return_when=asyncio.FIRST_COMPLETED
                )

                for completed_task in done:
                    result = await completed_task
                    if progress_callback:
                        progress_callback(result)
                    yield result

                # For each completed task, try to add a new one
                for _ in range(len(done)):
                    try:
                        task_func = next(tasks_iter)
                        task = asyncio.create_task(task_func())  # type: ignore
                        pending_tasks.add(task)
                    except StopIteration:
                        break

        except Exception:
            for task in pending_tasks:
                task.cancel()
            raise



"""Refactored core tests - more concise with single responsibility per test."""

import asyncio

# Removed Mock import - using real implementations
import pytest
from tenacity import AsyncRetrying, Retrying, stop_after_attempt

from dotevals import ForEach, foreach
from dotevals.concurrency import SlidingWindow
from dotevals.core import run_evaluation
from dotevals.evaluators import evaluator, exact_match
from dotevals.metrics import Metric, accuracy, metric, registry
from dotevals.models import Result, Score
from dotevals.sessions import SessionManager


@pytest.fixture
def simple_dataset():
    """Basic 3-item dataset for testing."""
    return [("a", 1), ("b", 2), ("c", 3)]


@pytest.fixture
def large_dataset():
    """100-item dataset for sampling tests."""
    return [(str(i), i) for i in range(100)]


@pytest.fixture
def session_manager(tmp_path):
    """Session manager with JSON storage."""
    import uuid

    unique_id = str(uuid.uuid4())[:8]
    return SessionManager(
        evaluation_name=f"test_eval_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}/evaluations",
    )


@metric
def mean():
    """Custom mean metric for testing."""

    def metric_fn(scores: list[float]) -> float:
        return sum(scores) / len(scores) if scores else 0.0

    return metric_fn


@metric
def metric_any():
    """Custom any metric for testing."""

    def metric_fn(scores: list[bool]) -> bool:
        return any(scores)

    return metric_fn


@pytest.fixture(autouse=True)
def register_test_metrics():
    """Register custom metrics for tests and clean up after."""
    # Register custom metrics
    registry["mean"] = mean()
    registry["metric_any"] = metric_any()
    yield
    # Clean up after test
    if "mean" in registry:
        del registry["mean"]
    if "metric_any" in registry:
        del registry["metric_any"]


@pytest.mark.parametrize("is_async", [False, True])
@pytest.mark.asyncio
async def test_basic_evaluation(simple_dataset, session_manager, is_async):
    """Test basic evaluation works for both sync and async."""
    if is_async:

        @foreach("text,number", simple_dataset)
        async def eval_fn(text, number):
            await asyncio.sleep(0)  # Simulate async work
            return Result(exact_match(text, text))
    else:

        @foreach("text,number", simple_dataset)
        def eval_fn(text, number):
            return Result(exact_match(text, text))

    summary = await eval_fn(session_manager, samples=None)

    assert summary.summary["exact_match"]["accuracy"] == 1.0
    assert len(summary.results) == 3  # Check results count instead


@pytest.mark.parametrize("is_async", [False, True])
@pytest.mark.parametrize("samples", [0, 5, 50, 200])
@pytest.mark.asyncio
async def test_sampling(large_dataset, session_manager, is_async, samples):
    """Test sampling parameter works correctly."""
    if is_async:

        @foreach("text,number", large_dataset)
        async def eval_fn(text, number):
            return Result(exact_match(int(text), number))
    else:

        @foreach("text,number", large_dataset)
        def eval_fn(text, number):
            return Result(exact_match(int(text), number))

    summary = await eval_fn(session_manager, samples=samples)

    expected_count = min(samples, 100) if samples > 0 else 0
    assert len(summary.results) == expected_count


@pytest.mark.parametrize(
    "dataset,columns,expected_calls",
    [
        # Single column with tuple
        ([("a",), ("b",)], "text", [{"text": "a"}, {"text": "b"}]),
        # Single column with direct values
        (["a", "b"], "text", [{"text": "a"}, {"text": "b"}]),
        # Multiple columns
        (
            [("a", 1), ("b", 2)],
            "text,num",
            [{"text": "a", "num": 1}, {"text": "b", "num": 2}],
        ),
        # Dict dataset
        ([{"x": 1}, {"x": 2}], "data", [{"data": {"x": 1}}, {"data": {"x": 2}}]),
        # Many columns (6 columns test)
        (
            [("a", "b", "c", "d", "e", "f")],
            "c1,c2,c3,c4,c5,c6",
            [{"c1": "a", "c2": "b", "c3": "c", "c4": "d", "c5": "e", "c6": "f"}],
        ),
    ],
)
@pytest.mark.asyncio
async def test_dataset_formats(dataset, columns, expected_calls, session_manager):
    """Test different dataset formats are handled correctly."""
    captured_args = []

    @foreach(columns, dataset)
    def eval_fn(**kwargs):
        captured_args.append(kwargs)
        return Result(Score("test", True, [accuracy()]))

    await eval_fn(session_manager, samples=None)

    assert len(captured_args) == len(expected_calls)
    for actual, expected in zip(captured_args, expected_calls):
        assert actual == expected


@pytest.mark.asyncio
async def test_empty_dataset(session_manager):
    """Test empty dataset returns empty summary."""

    @foreach("text", [])
    def eval_fn(text):
        return Result(exact_match(text, text))

    summary = await eval_fn(session_manager, samples=None)
    assert summary.summary == {}
    assert len(summary.results) == 0


@pytest.mark.asyncio
async def test_multiple_evaluators(simple_dataset, session_manager):
    """Test multiple evaluators in single result."""

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number):
        return Result(
            exact_match(text, text, name="text_match"),
            exact_match(number, number, name="number_match"),
            prompt=text,
        )

    summary = await eval_fn(session_manager, samples=None)

    assert "text_match" in summary.summary
    assert "number_match" in summary.summary
    assert summary.summary["text_match"]["accuracy"] == 1.0
    assert summary.summary["number_match"]["accuracy"] == 1.0


@pytest.mark.asyncio
async def test_multiple_metrics(simple_dataset, session_manager):
    """Test multiple metrics on single evaluator."""
    mean_metric = mean()

    @evaluator(metrics=[accuracy(), mean_metric])
    def value_eval(value, target):
        return value == target

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number):
        # Pass 1.0 for first item, 0.0 for second, 1.0 for third -> mean = 2/3
        return Result(value_eval(number, number if number != 2 else -1))

    summary = await eval_fn(session_manager, samples=None)

    assert summary.summary["value_eval"]["accuracy"] == pytest.approx(2 / 3)
    assert summary.summary["value_eval"]["mean"] == pytest.approx(2 / 3)


@pytest.mark.asyncio
async def test_custom_metric_any(simple_dataset, session_manager):
    """Test custom metric_any aggregation."""

    @evaluator(metrics=[accuracy(), metric_any()])
    def dummy_match(answer, target):
        return answer == target

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number):
        # Will match only first item where text=="a"
        return Result(dummy_match(text, "a"))

    summary = await eval_fn(session_manager, samples=None)

    assert summary.summary["dummy_match"]["accuracy"] == pytest.approx(1 / 3)
    assert summary.summary["dummy_match"]["metric_any"] is True


@pytest.mark.asyncio
async def test_evaluation_error_handling(simple_dataset, session_manager):
    """Test errors are captured in results."""
    call_count = 0

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number):
        nonlocal call_count
        call_count += 1
        if call_count == 2:
            raise ValueError("Test error")
        return Result(exact_match(text, text))

    summary = await eval_fn(session_manager, samples=None)

    # Check we have 3 total results: 2 successful and 1 error
    assert len(summary.results) == 3
    assert len([r for r in summary.results if not r.error]) == 2  # 2 successful results
    error_results = [r for r in summary.results if r.error]
    assert len(error_results) == 1
    assert "ValueError: Test error" in error_results[0].error

    # Accuracy should be calculated only on the 2 successful results (both are True)
    # Since only 2 items were evaluated successfully, accuracy = 2/3
    assert summary.summary["exact_match"]["accuracy"] == pytest.approx(2 / 3)


@pytest.mark.asyncio
async def test_sync_retry_configuration(tmp_path):
    """Test sync evaluation with custom retry configuration."""
    foreach_with_retry = ForEach(retries=Retrying(stop=stop_after_attempt(5)))

    attempt_count = 0

    @foreach_with_retry("text", [("a",)])
    def eval_fn(text):
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise ConnectionError("Retry me")
        return Result(exact_match(text, "a"))

    # Use real SessionManager with temporary storage
    import uuid

    unique_id = str(uuid.uuid4())[:8]
    session_manager = SessionManager(
        experiment_name=f"test_retry_{unique_id}",
        evaluation_name=f"test_eval_{unique_id}",
        storage=f"json://{tmp_path}/evaluations",
    )

    summary = await eval_fn(session_manager, samples=None)
    assert attempt_count == 3
    # Verify the result was successful after retries
    assert summary.summary["exact_match"]["accuracy"] == 1.0


@pytest.mark.asyncio
async def test_async_retry_configuration(tmp_path):
    """Test async evaluation with custom retry configuration."""
    foreach_with_retry = ForEach(retries=AsyncRetrying(stop=stop_after_attempt(5)))

    attempt_count = 0

    @foreach_with_retry("text", [("a",)])
    async def eval_fn(text):
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise ConnectionError("Retry me")
        return Result(exact_match(text, "a"))

    # Use real SessionManager with temporary storage
    import uuid

    unique_id = str(uuid.uuid4())[:8]
    session_manager = SessionManager(
        experiment_name=f"test_async_retry_{unique_id}",
        evaluation_name=f"test_eval_{unique_id}",
        storage=f"json://{tmp_path}/evaluations",
    )

    summary = await eval_fn(session_manager, samples=None)
    assert attempt_count == 3
    # Verify the result was successful after retries
    assert summary.summary["exact_match"]["accuracy"] == 1.0


@pytest.mark.asyncio
async def test_async_evaluation_without_retry(tmp_path):
    """Test async evaluation without retry on exceptions."""
    from tenacity import AsyncRetrying, stop_after_attempt

    # Create a no-retry configuration (single attempt only)
    no_retry = AsyncRetrying(stop=stop_after_attempt(1))
    foreach_no_retry = ForEach(retries=no_retry)

    @foreach_no_retry("text", [("a",)])
    async def eval_fn(text):
        raise ConnectionError("Should not retry")

    # Use real SessionManager with temporary storage
    import uuid

    unique_id = str(uuid.uuid4())[:8]
    session_manager = SessionManager(
        experiment_name=f"test_no_retry_{unique_id}",
        evaluation_name=f"test_eval_{unique_id}",
        storage=f"json://{tmp_path}/evaluations",
    )

    # Should complete without retrying (error captured in result)
    summary = await eval_fn(session_manager, samples=None)
    # Verify error was captured in results
    assert len(summary.results) == 1
    assert summary.results[0].error is not None
    assert "ConnectionError" in summary.results[0].error


@pytest.mark.asyncio
async def test_async_evaluation_with_exceptions(tmp_path):
    """Test async evaluation handles exceptions properly."""

    @foreach("text", [("a",), ("b",), ("c",)])
    async def eval_fn(text):
        if text == "b":
            raise RuntimeError("Expected error")
        return Result(exact_match(text, text))

    # Use real SessionManager with temporary storage
    import uuid

    unique_id = str(uuid.uuid4())[:8]
    session_manager = SessionManager(
        experiment_name=f"test_exceptions_{unique_id}",
        evaluation_name=f"test_eval_{unique_id}",
        storage=f"json://{tmp_path}/evaluations",
    )

    summary = await eval_fn(session_manager, samples=None)
    # Should have processed all 3 items (2 success, 1 error)
    assert len(summary.results) == 3
    # Check that we have 2 successful results and 1 error
    errors = [r for r in summary.results if r.error is not None]
    successes = [r for r in summary.results if r.error is None]
    assert len(errors) == 1
    assert len(successes) == 2
    assert "RuntimeError" in errors[0].error


@pytest.mark.asyncio
async def test_additional_kwargs_passed_through(simple_dataset, session_manager):
    """Test additional kwargs are passed to evaluation function."""
    captured_kwargs = []

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number, custom_param=None, another_param=42):
        captured_kwargs.append(
            {"custom_param": custom_param, "another_param": another_param}
        )
        return Result(exact_match(text, text))

    await eval_fn(
        session_manager, samples=None, custom_param="test_value", another_param=100
    )

    assert len(captured_kwargs) == 3
    for kwargs in captured_kwargs:
        assert kwargs["custom_param"] == "test_value"
        assert kwargs["another_param"] == 100


@pytest.mark.asyncio
async def test_sample_order_preserved_sync(large_dataset, session_manager):
    """Test sync evaluation preserves dataset order."""
    processed_order = []

    @foreach("text,number", large_dataset)
    def eval_fn(text, number):
        processed_order.append(number)
        return Result(exact_match(True, True))

    await eval_fn(session_manager, samples=10)

    assert processed_order == list(range(10))


@pytest.mark.asyncio
async def test_sample_order_may_vary_async(large_dataset, session_manager):
    """Test async evaluation may process out of order."""
    processed_order = []

    @foreach("text,number", large_dataset)
    async def eval_fn(text, number):
        # Add variable delay to encourage out-of-order processing
        await asyncio.sleep(0.001 * (10 - number) if number < 10 else 0)
        processed_order.append(number)
        return Result(exact_match(True, True))

    await eval_fn(session_manager, samples=10)

    assert len(processed_order) == 10
    assert set(processed_order) == set(range(10))


@pytest.mark.asyncio
async def test_run_evaluation_directly_with_samples(large_dataset, session_manager):
    """Test calling run_evaluation directly with samples parameter."""

    def eval_fn(text, number):
        return Result(exact_match(int(text), number))

    summary = await run_evaluation(
        eval_fn, "text,number", large_dataset, session_manager, samples=5
    )

    assert len(summary.results) == 5


@pytest.mark.asyncio
async def test_evaluation_with_session_manager(simple_dataset, tmp_path):
    """Test evaluation with explicit session manager configuration."""
    session_mgr = SessionManager(
        storage=f"json://{tmp_path}",
        experiment_name="test_exp",
        evaluation_name="test_eval",
    )

    @foreach("text,number", simple_dataset)
    def eval_fn(text, number):
        return Result(exact_match(text, text))

    result = await eval_fn(session_manager=session_mgr, samples=None)

    assert isinstance(result.summary, dict)
    assert result.summary["exact_match"]["accuracy"] == 1.0


@pytest.mark.asyncio
async def test_async_evaluation_with_session_manager(simple_dataset, tmp_path):
    """Test async evaluation with explicit session manager configuration."""
    session_mgr = SessionManager(
        storage=f"json://{tmp_path}",
        experiment_name="test_exp",
        evaluation_name="test_eval",
    )

    @foreach("text,number", simple_dataset)
    async def eval_fn(text, number):
        await asyncio.sleep(0.001)
        return Result(exact_match(text, text))

    result = await eval_fn(session_manager=session_mgr, samples=None)

    assert isinstance(result.summary, dict)
    assert result.summary["exact_match"]["accuracy"] == 1.0


@pytest.mark.asyncio
async def test_async_evaluator_partial_failure(tmp_path):
    """Test handling partial failures in async evaluation."""
    import uuid

    from dotevals.sessions import SessionManager

    unique_id = str(uuid.uuid4())[:8]
    session_mgr = SessionManager(
        evaluation_name=f"test_eval_{unique_id}",
        experiment_name=f"test_exp_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    call_count = 0

    async def sometimes_failing(input):  # Parameter name must match column name
        nonlocal call_count
        call_count += 1
        if call_count == 2:
            raise ValueError(f"Failed on item {call_count}")
        return Result(Score("test", True, []))

    dataset = [{"input": f"test_{i}"} for i in range(3)]

    # Errors are captured in results, not propagated
    from dotevals.concurrency import AsyncSequential

    summary = await run_evaluation(
        sometimes_failing,
        "input",  # column_spec should be a string
        dataset,
        session_manager=session_mgr,
        concurrency=AsyncSequential(),  # Use AsyncSequential for async evaluator
    )

    # Check that we have 2 successful results and 1 error
    assert len(summary.results) == 3
    successful = [r for r in summary.results if not r.error]
    errors = [r for r in summary.results if r.error]

    assert len(successful) == 2
    assert len(errors) == 1

    # The second item should have failed
    assert errors[0].dataset_row == {"input": {"input": "test_1"}}
    assert "ValueError" in errors[0].error
    assert "Failed on item 2" in errors[0].error


@pytest.mark.asyncio
async def test_async_concurrent_evaluation_errors(tmp_path):
    """Test error handling with concurrent async evaluation."""
    import uuid

    from dotevals.sessions import SessionManager

    unique_id = str(uuid.uuid4())[:8]
    session_mgr = SessionManager(
        experiment_name=f"test_exp_{unique_id}",
        evaluation_name=f"test_eval_{unique_id}",
        storage=f"json://{tmp_path}",
    )

    async def evaluator_with_errors(id):  # Parameter name must match column name
        if id % 3 == 0:
            raise ValueError(f"Error on {id}")
        await asyncio.sleep(0.01)
        return Result(Score("test", id, []))

    dataset = [(i,) for i in range(10)]  # Use tuples for single column

    # Errors are captured in results, not propagated
    from dotevals.concurrency import Adaptive

    summary = await run_evaluation(
        evaluator_with_errors,
        "id",  # column_spec should be a string
        dataset,
        session_manager=session_mgr,
        concurrency=Adaptive(initial_concurrency=3),  # Use Adaptive strategy for async
    )

    # Check that we have both successful and error results
    successful = [r for r in summary.results if not r.error]
    errors = [r for r in summary.results if r.error]

    # IDs 0, 3, 6, 9 should have errors (multiples of 3)
    assert len(errors) == 4
    assert len(successful) == 6

    # Check error messages
    for error_result in errors:
        assert "ValueError" in error_result.error
        assert "Error on" in error_result.error



from .core import ForEach as ForEach
from .core import foreach as foreach
from .models import Result as Result
from .progress import MultiProgress, ProgressTracker, SingleProgress
from .providers import ModelHandle as ModelHandle
from .providers import ModelProvider as ModelProvider
from .runners import Runner as Runner



"""
Shared utility functions for dotevals.

This module contains common functionality used across multiple components,
particularly serialization logic shared between storage backends.
"""

import base64
import io
from collections.abc import Callable
from typing import Any

from PIL import Image

from dotevals.models import Record, Result, Score


def recursive_map(obj: Any, transform_func: Callable[[Any], Any]) -> Any:
    """
    Recursively apply a transformation function to all values in a nested structure.

    This function traverses through dictionaries, lists, sets, and tuples,
    applying the transform_func to each value it encounters.

    Args:
        obj: The object to transform
        transform_func: Function to apply to each value

    Returns:
        The object with all values transformed by transform_func
    """
    if isinstance(obj, dict):
        return {k: recursive_map(v, transform_func) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [recursive_map(x, transform_func) for x in obj]
    elif isinstance(obj, set):
        return {recursive_map(x, transform_func) for x in obj}
    elif isinstance(obj, tuple):
        return tuple(recursive_map(x, transform_func) for x in obj)
    else:
        return transform_func(obj)


def serialize_value(value: Any) -> Any:
    """Recursively serialize values, converting PIL Images to base64.

    This function handles the conversion of complex Python objects into
    JSON-serializable formats, with special handling for PIL Images.

    Args:
        value: The value to serialize

    Returns:
        A JSON-serializable representation of the value
    """
    if isinstance(value, Image.Image):
        # Convert PIL Image to base64
        buffered = io.BytesIO()
        value.save(buffered, format="PNG")
        img_base64 = base64.b64encode(buffered.getvalue()).decode()
        return {
            "__type__": "PIL.Image",
            "data": img_base64,
            "mode": value.mode,
            "size": value.size,
        }
    elif isinstance(value, dict):
        return {k: serialize_value(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [serialize_value(item) for item in value]
    elif isinstance(value, tuple):
        return {"__type__": "tuple", "data": [serialize_value(item) for item in value]}
    elif isinstance(value, set):
        return {"__type__": "set", "data": [serialize_value(item) for item in value]}
    else:
        return value


def deserialize_value(value: Any) -> Any:
    """Recursively deserialize values, converting base64 back to PIL Images.

    This function reverses the serialization process, reconstructing
    complex Python objects from their JSON-serializable representations.

    Args:
        value: The serialized value to deserialize

    Returns:
        The original Python object
    """
    if isinstance(value, dict):
        if value.get("__type__") == "PIL.Image":
            # Convert base64 back to PIL Image
            img_data = base64.b64decode(value["data"])
            img = Image.open(io.BytesIO(img_data))
            return img
        elif value.get("__type__") == "tuple":
            return tuple(deserialize_value(item) for item in value["data"])
        elif value.get("__type__") == "set":
            return {deserialize_value(item) for item in value["data"]}
        else:
            return {k: deserialize_value(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [deserialize_value(item) for item in value]
    else:
        return value


def serialize_records(results: list[Record]) -> list[dict]:
    """Serialize a list of Record objects to JSON-compatible dictionaries.

    Args:
        results: List of Record objects to serialize

    Returns:
        List of dictionaries representing the serialized records
    """
    data = [
        {
            "item_id": r.item_id,
            "result": {
                "prompt": r.result.prompt,
                "scores": [
                    {
                        "name": s.name,
                        "value": s.value,
                        "metrics": [metric.__name__ for metric in s.metrics],
                        "metadata": s.metadata,
                    }
                    for s in r.result.scores
                ],
                "error": r.result.error,
                "model_response": r.result.model_response,
            },
            "dataset_row": serialize_value(r.dataset_row),
            "error": r.error,
            "timestamp": r.timestamp,
        }
        for r in results
    ]

    return data


def deserialize_records(data: list[dict]) -> list[Record]:
    """Deserialize a list of dictionaries back to Record objects.

    Args:
        data: List of dictionaries representing serialized records

    Returns:
        List of Record objects
    """
    from dotevals.metrics import registry

    results = []
    for r_data in data:
        scores = [
            Score(
                name=s_data["name"],
                value=s_data["value"],
                metrics=[registry[metric_name] for metric_name in s_data["metrics"]],
                metadata=s_data.get("metadata", {}),
            )
            for s_data in r_data["result"]["scores"]
        ]

        result = Result(
            *scores,
            prompt=r_data["result"].get("prompt"),
            error=r_data["result"].get("error"),
            model_response=r_data["result"].get("model_response"),
        )

        record = Record(
            result=result,
            item_id=r_data["item_id"],
            dataset_row=deserialize_value(r_data["dataset_row"]),
            error=r_data.get("error"),
            timestamp=r_data.get("timestamp") or 0,
        )

        results.append(record)

    return results



from abc import ABC, abstractmethod
from typing import Optional

from dotevals.models import Evaluation, EvaluationStatus, Record

__all__ = ["Storage", "StorageRegistry", "registry"]


class Storage(ABC):
    """Abstract storage interface"""

    @abstractmethod
    def create_experiment(self, experiment_name: str) -> None:
        """Create an experiment. Should be idempotent - if experiment exists, do nothing."""
        pass

    @abstractmethod
    def delete_experiment(self, experiment_name: str) -> None:
        pass

    @abstractmethod
    def rename_experiment(self, old_name: str, new_name: str) -> None:
        pass

    @abstractmethod
    def list_experiments(self) -> list[str]:
        pass

    @abstractmethod
    def create_evaluation(self, experiment_name: str, evaluation: Evaluation) -> None:
        pass

    @abstractmethod
    def load_evaluation(
        self, experiment_name: str, evaluation_name: str
    ) -> Evaluation | None:
        pass

    @abstractmethod
    def update_evaluation_status(
        self, experiment_name: str, evaluation_name: str, status: EvaluationStatus
    ) -> None:
        pass

    @abstractmethod
    def completed_items(self, experiment_name: str, evaluation_name: str) -> list[int]:
        pass

    @abstractmethod
    def list_evaluations(self, experiment_name: str) -> list[str]:
        pass

    @abstractmethod
    def add_results(
        self,
        experiment_name: str,
        evaluation_name: str,
        results: list[Record],
    ) -> None:
        pass

    @abstractmethod
    def get_results(self, experiment_name: str, evaluation_name: str) -> list[Record]:
        pass

    @abstractmethod
    def remove_error_result(
        self, experiment_name: str, evaluation_name: str, item_id: int
    ) -> None:
        """Remove an errored result for a specific item that will be retried."""
        pass

    def remove_error_results_batch(
        self, experiment_name: str, evaluation_name: str, item_ids: list[int]
    ) -> None:
        """Remove multiple errored results in a batch.

        Default implementation calls remove_error_result for each item.
        Storage backends should override this for better performance.
        """
        for item_id in item_ids:
            self.remove_error_result(experiment_name, evaluation_name, item_id)


class StorageRegistry:
    """Registry for storage backends."""

    def __init__(self):
        self._backends = {}
        self._plugins_loaded = False

        # Register built-in backends
        from .json import JSONStorage

        self.register("json", JSONStorage)

    def register(self, name: str, storage_class: type[Storage]) -> None:
        """Register a storage backend.

        Args:
            name: The name of the backend (e.g., "json", "sqlite", "redis")
            storage_class: The storage class that implements the Storage interface
        """
        self._backends[name] = storage_class

    def get_backend(self, name: str) -> type[Storage]:
        """Get a storage backend by name.

        Args:
            name: The name of the backend

        Returns:
            The storage class

        Raises:
            ValueError: If the backend is not registered
        """
        if name not in self._backends:
            raise ValueError(f"Unknown storage backend: {name}")
        return self._backends[name]

    def get_storage(self, storage_path: str | None = None) -> Storage:
        """Get a storage instance based on the storage path.

        The storage path format is: backend://path
        For backward compatibility, paths without a backend default to JSON.

        Args:
            storage_path: The storage path (e.g., "json://data", "sqlite://db.sqlite")
                         If None, defaults to "json://.dotevals"

        Returns:
            A storage instance

        Raises:
            ValueError: If the backend is not registered
        """
        # Ensure plugins are loaded
        self.load_plugins()

        storage_path = "json://.dotevals" if storage_path is None else storage_path

        if "://" in storage_path:
            backend_name, path = storage_path.split("://", 1)
        else:
            backend_name, path = "json", storage_path

        backend_class = self.get_backend(backend_name)

        return backend_class(path)  # type: ignore[call-arg]

    def list_backends(self) -> list[str]:
        """List all registered backend names."""
        # Ensure plugins are loaded
        self.load_plugins()
        return list(self._backends.keys())

    def load_plugins(self, force: bool = False):
        """Load storage backends from entry points.

        Args:
            force: If True, reload plugins even if already loaded
        """
        if self._plugins_loaded and not force:
            return

        try:
            import importlib.metadata

            # Load all entry points in the dotevals.storage group
            entry_points = importlib.metadata.entry_points()

            # Handle different Python versions
            storage_entries = []
            if hasattr(entry_points, "select"):
                # Python 3.10+
                storage_entries = list(entry_points.select(group="dotevals.storage"))
            else:
                # Python 3.9 - entry_points is a dict-like object
                group = entry_points.get("dotevals.storage")
                if group is not None:
                    storage_entries = list(group)

            for entry_point in storage_entries:
                try:
                    # Load the storage class
                    storage_class = entry_point.load()
                    # Register it with the name from entry point
                    self.register(entry_point.name, storage_class)
                except Exception as e:
                    # Log but don't fail if a plugin can't be loaded
                    import warnings

                    warnings.warn(
                        f"Failed to load storage plugin '{entry_point.name}': {e}",
                        RuntimeWarning,
                    )
        except ImportError:
            # importlib.metadata not available (shouldn't happen in Python 3.8+)
            pass

        self._plugins_loaded = True


# Global registry instance
registry = StorageRegistry()



"""Test the batch execution strategy."""

from dotevals.concurrency import Batch


def test_batch_execution():
    """Test that tasks are executed in batches."""
    strategy = Batch(batch_size=3)
    execution_times = []

    def create_tasks():
        for i in range(7):

            def task(task_id=i):
                execution_times.append(task_id)
                return f"result_{task_id}"

            yield task

    results = list(strategy.execute(create_tasks()))

    # Check all tasks executed
    assert len(execution_times) == 7
    assert len(results) == 7
    assert results == [f"result_{i}" for i in range(7)]


def test_batch_with_progress_callback():
    """Test batch execution with progress callback."""
    strategy = Batch(batch_size=2)
    progress_results = []

    def progress_callback(result):
        progress_results.append(result)

    def create_tasks():
        for i in range(5):

            def task(task_id=i):
                return f"result_{task_id}"

            yield task

    results = list(strategy.execute(create_tasks(), progress_callback))

    assert len(results) == 5
    assert len(progress_results) == 5
    assert progress_results == [f"result_{i}" for i in range(5)]



"""Comprehensive CLI tests combining real implementations and targeted mocking.

Tests are organized by functionality:
- Real implementation tests for integration testing
- Mocked tests only where necessary for formatting, error simulation, etc.
"""

import json
import tempfile
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner
from rich.console import Console

from dotevals.cli import cli
from dotevals.datasets import Dataset
from dotevals.datasets import _registry as dataset_registry
from dotevals.metrics import accuracy
from dotevals.models import Evaluation, EvaluationStatus, Record, Result, Score
from dotevals.storage import registry as storage_registry
from dotevals.storage.base import Storage
from dotevals.storage.json import JSONStorage


class TestCLICommands:
    """Test CLI commands with real implementations."""

    @pytest.fixture(autouse=True)
    def clean_registries(self):
        """Clean up registries before and after each test."""
        # Store original registries
        original_storage = storage_registry._backends.copy()
        original_datasets = dataset_registry._dataset_classes.copy()
        original_discovery = dataset_registry._discovery_completed

        yield

        # Restore original registries
        storage_registry._backends = original_storage
        dataset_registry._dataset_classes = original_datasets
        dataset_registry._discovery_completed = original_discovery

    def test_list_command(self):
        """Test list command with real JSON storage backend."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create real storage and data
            storage = JSONStorage(tmpdir)
            storage.create_experiment("real_project")
            storage.create_experiment("another_project")
            storage.create_experiment("20240101_120000_abcd1234")  # ephemeral
            storage.create_experiment(".dotevals")  # Should be skipped

            # Add evaluations with real data
            eval1 = Evaluation(
                evaluation_name="real_eval",
                started_at=1234567890.0,
                status=EvaluationStatus.COMPLETED,
            )
            storage.create_evaluation("real_project", eval1)

            # Use the actual CLI
            runner = CliRunner()
            result = runner.invoke(cli, ["list", "--storage", f"json://{tmpdir}"])

            # Verify output contains our experiments
            assert result.exit_code == 0
            assert "real_project" in result.output
            assert "another_project" in result.output
            # Check for ephemeral experiments section
            assert (
                "20240101_120000_abcd1234" in result.output
                or "Ephemeral" in result.output
            )

            # Test with name filter
            result = runner.invoke(
                cli, ["list", "--name", "real", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "real_project" in result.output
            assert "another_project" not in result.output

            # Test empty storage
            with tempfile.TemporaryDirectory() as empty_dir:
                result = runner.invoke(
                    cli, ["list", "--storage", f"json://{empty_dir}"]
                )
                assert result.exit_code == 0
                assert "No experiments found" in result.output

    def test_show_command(self):
        """Test show command with real storage and evaluation data."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)
            storage.create_experiment("test_exp")

            # Create real evaluation
            evaluation = Evaluation(
                evaluation_name="test_eval",
                started_at=1234567890.0,
                status=EvaluationStatus.COMPLETED,
                metadata={"model": "test-model", "temperature": 0.5},
            )
            storage.create_evaluation("test_exp", evaluation)

            # Add real results
            records = [
                Record(
                    Result(
                        Score("exact_match", True, [accuracy()], {}),
                        prompt="Test prompt 1",
                    ),
                    0,
                    {"input": "2+2", "expected": "4", "actual": "4"},
                ),
                Record(
                    Result(
                        Score("exact_match", False, [accuracy()], {}),
                        prompt="Test prompt 2",
                    ),
                    1,
                    {"input": "3+3", "expected": "6", "actual": "5"},
                ),
                Record(
                    Result(
                        Score("exact_match", True, [accuracy()], {}),
                        prompt="Test prompt 3",
                    ),
                    2,
                    {"input": "4+4", "expected": "8", "actual": "8"},
                    error="Connection timeout",
                ),
            ]
            storage.add_results("test_exp", "test_eval", records)

            runner = CliRunner()

            # Test summary view
            result = runner.invoke(
                cli, ["show", "test_exp", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "test_eval" in result.output
            assert "exact_match" in result.output
            assert "accuracy" in result.output

            # Test full JSON output
            result = runner.invoke(
                cli, ["show", "test_exp", "--full", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            # Should contain JSON data
            assert '"prompt"' in result.output
            assert '"scores"' in result.output

            # Test error details
            result = runner.invoke(
                cli, ["show", "test_exp", "--errors", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "Connection timeout" in result.output

            # Test specific evaluation
            result = runner.invoke(
                cli,
                [
                    "show",
                    "test_exp",
                    "--evaluation",
                    "test_eval",
                    "--storage",
                    f"json://{tmpdir}",
                ],
            )
            assert result.exit_code == 0
            assert "test_eval" in result.output

            # Test non-existent evaluation
            result = runner.invoke(
                cli,
                [
                    "show",
                    "test_exp",
                    "--evaluation",
                    "nonexistent",
                    "--storage",
                    f"json://{tmpdir}",
                ],
            )
            assert result.exit_code == 0
            assert "not found" in result.output.lower()

            # Test experiment with no evaluations
            storage.create_experiment("empty_exp")
            result = runner.invoke(
                cli, ["show", "empty_exp", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "No evaluations found" in result.output

            # Test with all results having errors (covers line 195)
            storage.create_experiment("all_errors_exp")
            eval_err = Evaluation(
                evaluation_name="error_eval",
                started_at=1234567890.0,
                status=EvaluationStatus.COMPLETED,
            )
            storage.create_evaluation("all_errors_exp", eval_err)
            error_records = [
                Record(
                    Result(Score("exact_match", False, [accuracy()], {})),
                    i,
                    {"input": f"test_{i}"},
                    error=f"Error {i}",
                )
                for i in range(3)
            ]
            storage.add_results("all_errors_exp", "error_eval", error_records)
            result = runner.invoke(
                cli, ["show", "all_errors_exp", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            # Should show error summary
            assert "Error Summary" in result.output or "errors" in result.output.lower()

    def test_rename_command(self):
        """Test rename command with real storage operations."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)
            storage.create_experiment("old_name")

            # Add some data to ensure it's preserved
            eval1 = Evaluation(
                evaluation_name="eval1",
                started_at=1234567890.0,
                status=EvaluationStatus.COMPLETED,
            )
            storage.create_evaluation("old_name", eval1)

            runner = CliRunner()

            # Rename the experiment
            result = runner.invoke(
                cli, ["rename", "old_name", "new_name", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0

            # Verify rename worked
            experiments = storage.list_experiments()
            assert "new_name" in experiments
            assert "old_name" not in experiments

            # Verify data was preserved
            evaluations = storage.list_evaluations("new_name")
            assert "eval1" in evaluations

    def test_delete_command(self):
        """Test delete command with real storage operations."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)
            storage.create_experiment("to_delete")
            storage.create_experiment("to_keep")

            runner = CliRunner()

            # Delete one experiment
            result = runner.invoke(
                cli, ["delete", "to_delete", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0

            # Verify deletion
            experiments = storage.list_experiments()
            assert "to_delete" not in experiments
            assert "to_keep" in experiments

            # Try to delete non-existent
            result = runner.invoke(
                cli, ["delete", "nonexistent", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "not found" in result.output

    def test_datasets_command(self):
        """Test datasets command with real dataset implementations."""

        # Create real dataset classes
        class TestDataset1(Dataset):
            name = "test_dataset1"
            splits = ["train", "test", "validation"]
            columns = ["question", "answer", "explanation"]
            num_rows = 1000

            def __init__(self, split=None, **kwargs):
                self.split = split

            def __iter__(self):
                # Return a few sample items
                for i in range(3):
                    yield (f"Question {i}", f"Answer {i}", f"Explanation {i}")

        class TestDataset2(Dataset):
            name = "test_dataset2"
            splits = []  # No splits
            columns = ["input", "output"]
            num_rows = None  # Streaming dataset

            def __init__(self, split=None, **kwargs):
                pass

            def __iter__(self):
                for i in range(2):
                    yield (f"Input {i}", f"Output {i}")

        class BrokenDataset(Dataset):
            name = "broken_dataset"
            splits = ["train"]
            columns = ["data"]

            def __init__(self, split=None, **kwargs):
                raise ValueError("This dataset is broken")

            def __iter__(self):
                yield ("never called",)

        # Register the datasets
        dataset_registry.register(TestDataset1)
        dataset_registry.register(TestDataset2)
        dataset_registry.register(BrokenDataset)

        runner = CliRunner()

        # Test basic listing
        result = runner.invoke(cli, ["datasets"])
        assert result.exit_code == 0
        assert "test_dataset1" in result.output
        assert "test_dataset2" in result.output
        assert "train, test, validation" in result.output or "train" in result.output

        # Test verbose mode
        result = runner.invoke(cli, ["datasets", "--verbose"])
        assert result.exit_code == 0
        assert "test_dataset1" in result.output
        assert "test_dataset2" in result.output
        assert (
            "question, answer, explanation" in result.output
            or "question" in result.output
        )
        assert "@foreach.test_dataset1" in result.output
        assert "@foreach.test_dataset2()" in result.output  # No splits version
        # Broken dataset should be listed (error only happens on instantiation)
        assert "broken_dataset" in result.output

        # Test filtering
        result = runner.invoke(cli, ["datasets", "--name", "dataset1"])
        assert result.exit_code == 0
        assert "test_dataset1" in result.output
        assert "test_dataset2" not in result.output

        # Test filter with no matches
        result = runner.invoke(cli, ["datasets", "--name", "nonexistent"])
        assert result.exit_code == 0
        assert "No datasets found matching" in result.output

        # Test when no datasets available
        # Clear registry
        dataset_registry._dataset_classes.clear()
        result = runner.invoke(cli, ["datasets"])
        assert result.exit_code == 0
        assert "No datasets found" in result.output or "pip install" in result.output

    def test_error_handling(self):
        """Test error scenarios with real storage backend."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)

            runner = CliRunner()

            # Show non-existent experiment
            result = runner.invoke(
                cli, ["show", "nonexistent", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "not found" in result.output.lower()

            # Rename non-existent experiment
            result = runner.invoke(
                cli, ["rename", "nonexistent", "new", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "not found" in result.output.lower()

            # Create experiment and try to rename to existing name
            storage.create_experiment("exp1")
            storage.create_experiment("exp2")
            result = runner.invoke(
                cli, ["rename", "exp1", "exp2", "--storage", f"json://{tmpdir}"]
            )
            assert result.exit_code == 0
            assert "already exists" in result.output.lower()

            # Test delete with exception handling
            with patch(
                "dotevals.storage.json.JSONStorage.delete_experiment",
                side_effect=Exception("Storage error"),
            ):
                result = runner.invoke(
                    cli, ["delete", "exp1", "--storage", f"json://{tmpdir}"]
                )
                assert result.exit_code == 0
                assert "Error deleting experiment" in result.output

            # Test datasets command with exception
            with patch(
                "dotevals.cli.list_available", side_effect=Exception("Dataset error")
            ):
                result = runner.invoke(cli, ["datasets"])
                assert result.exit_code == 0
                assert "Error listing datasets" in result.output

    def test_custom_storage_plugin(self):
        """Test that custom storage plugins work with real registration."""

        # Create a real custom storage implementation
        class CustomStorage(Storage):
            def __init__(self, path):
                self.path = path
                self.experiments = {
                    "custom_exp1": ["eval1", "eval2"],
                    "custom_exp2": ["eval3"],
                }

            def list_experiments(self):
                return list(self.experiments.keys())

            def list_evaluations(self, experiment_name):
                return self.experiments.get(experiment_name, [])

            def get_results(self, experiment_name, evaluation_name):
                # Return some dummy results
                return [
                    Record(
                        Result(Score("custom_metric", True, [accuracy()], {})),
                        0,
                        {"data": "custom"},
                    )
                ]

            def create_experiment(self, name):
                self.experiments[name] = []

            def delete_experiment(self, name):
                self.experiments.pop(name, None)

            def rename_experiment(self, old_name, new_name):
                if old_name in self.experiments:
                    self.experiments[new_name] = self.experiments.pop(old_name)

            def create_evaluation(self, experiment_name, evaluation):
                if experiment_name not in self.experiments:
                    self.experiments[experiment_name] = []
                self.experiments[experiment_name].append(evaluation.evaluation_name)

            def load_evaluation(self, experiment_name, evaluation_name):
                if evaluation_name in self.experiments.get(experiment_name, []):
                    return Evaluation(
                        evaluation_name=evaluation_name,
                        started_at=1234567890.0,
                        status=EvaluationStatus.COMPLETED,
                    )
                return None

            def update_evaluation_status(
                self, experiment_name, evaluation_name, status
            ):
                pass  # No-op for this test

            def add_results(self, experiment_name, evaluation_name, records):
                pass  # No-op for this test

            def completed_items(self, experiment_name, evaluation_name):
                return []

            def remove_error_result(self, experiment_name, evaluation_name, item_id):
                pass

            def remove_error_results_batch(
                self, experiment_name, evaluation_name, item_ids
            ):
                pass

        # Register the custom storage
        storage_registry.register("custom", CustomStorage)

        runner = CliRunner()

        # Use the custom storage with CLI
        result = runner.invoke(cli, ["list", "--storage", "custom://test_path"])
        assert result.exit_code == 0
        assert "custom_exp1" in result.output
        assert "custom_exp2" in result.output

        # Show command with custom storage
        result = runner.invoke(
            cli, ["show", "custom_exp1", "--storage", "custom://test_path"]
        )
        assert result.exit_code == 0
        assert "eval1" in result.output or "custom_metric" in result.output

    def test_multiple_dataset_plugins(self):
        """Test aggregating datasets from multiple plugins with real implementations."""

        # Simulate multiple dataset "plugins" by registering datasets with different prefixes
        class Plugin1Dataset1(Dataset):
            name = "plugin1_dataset1"
            splits = ["train", "test"]
            columns = ["col1", "col2"]
            num_rows = 100

            def __init__(self, split=None, **kwargs):
                self.split = split

            def __iter__(self):
                yield ("data1", "data2")

        class Plugin1Dataset2(Dataset):
            name = "plugin1_dataset2"
            splits = ["train"]
            columns = ["col3"]
            num_rows = 50

            def __init__(self, split=None, **kwargs):
                self.split = split

            def __iter__(self):
                yield ("data3",)

        class Plugin2Dataset1(Dataset):
            name = "plugin2_dataset1"
            splits = []
            columns = ["col4", "col5"]
            num_rows = None

            def __init__(self, split=None, **kwargs):
                pass

            def __iter__(self):
                yield ("data4", "data5")

        # Register all datasets
        dataset_registry.register(Plugin1Dataset1)
        dataset_registry.register(Plugin1Dataset2)
        dataset_registry.register(Plugin2Dataset1)

        runner = CliRunner()

        # List all datasets - should aggregate from all "plugins"
        result = runner.invoke(cli, ["datasets"])
        assert result.exit_code == 0
        assert "plugin1_dataset1" in result.output
        assert "plugin1_dataset2" in result.output
        assert "plugin2_dataset1" in result.output

        # Filter across plugins
        result = runner.invoke(cli, ["datasets", "--name", "dataset1"])
        assert result.exit_code == 0
        assert "plugin1_dataset1" in result.output
        assert "plugin2_dataset1" in result.output
        assert "plugin1_dataset2" not in result.output  # Doesn't match filter


class TestCLIFormatting:
    """Test Rich console formatting - these require mocking."""

    def test_colored_output_warnings(self):
        """Test that CLI uses proper coloring for warnings."""
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmpdir:
            captured_prints = []

            def capture_print(content=None, *args, **kwargs):
                if content is not None:
                    captured_prints.append(str(content))

            with patch("dotevals.cli.console.print", side_effect=capture_print):
                result = runner.invoke(cli, ["list", "--storage", f"json://{tmpdir}"])
                assert result.exit_code == 0

                # Check for yellow warning color
                assert any("[yellow]" in str(p) for p in captured_prints)
                assert any("No experiments found" in str(p) for p in captured_prints)

    def test_colored_output_errors(self):
        """Test that CLI uses red color for errors."""
        runner = CliRunner()

        captured_prints = []

        def capture_print(content=None, *args, **kwargs):
            if content is not None:
                captured_prints.append(str(content))

        with tempfile.TemporaryDirectory() as tmpdir:
            with patch("dotevals.cli.console.print", side_effect=capture_print):
                result = runner.invoke(
                    cli, ["show", "nonexistent", "--storage", f"json://{tmpdir}"]
                )
                assert result.exit_code == 0

                # Check for red error color
                assert any("[red]" in str(p) for p in captured_prints)
                assert any("not found" in str(p).lower() for p in captured_prints)

    def test_colored_output_success(self):
        """Test that CLI uses green color for success."""
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)
            storage.create_experiment("old_name")

            captured_prints = []

            def capture_print(content=None, *args, **kwargs):
                if content is not None:
                    captured_prints.append(str(content))

            with patch("dotevals.cli.console.print", side_effect=capture_print):
                result = runner.invoke(
                    cli,
                    ["rename", "old_name", "new_name", "--storage", f"json://{tmpdir}"],
                )
                assert result.exit_code == 0

                # Check for green success color
                assert any("[green]" in str(p) for p in captured_prints)
                assert any("renamed" in str(p).lower() for p in captured_prints)

    def test_rich_table_creation(self):
        """Test that CLI creates proper Rich tables."""
        runner = CliRunner()

        with tempfile.TemporaryDirectory() as tmpdir:
            storage = JSONStorage(tmpdir)
            storage.create_experiment("project1")
            storage.create_experiment("20240101_120000_abcd1234")  # ephemeral

            captured_tables = []

            def capture_print(content=None, *args, **kwargs):
                if content is not None and hasattr(content, "add_column"):
                    captured_tables.append(content)

            with patch("dotevals.cli.console.print", side_effect=capture_print):
                result = runner.invoke(cli, ["list", "--storage", f"json://{tmpdir}"])
                assert result.exit_code == 0

                # Should create tables for named and ephemeral experiments
                assert len(captured_tables) >= 1
                # Check table has proper structure
                table = captured_tables[0]
                assert hasattr(table, "title")
                assert hasattr(table, "columns")



import re
import tempfile
import time
from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import pytest

from dotevals.models import Evaluation, EvaluationStatus, Record, Result, Score
from dotevals.sessions import SessionManager, get_git_commit
from dotevals.storage.json import JSONStorage


@pytest.fixture
def temp_storage_dir(tmp_path):
    """Create a temporary directory for storage tests."""
    return tmp_path


def test_session_manager_init_without_experiment():
    """Test SessionManager initialization without experiment creates ephemeral experiment."""
    manager = SessionManager("eval_test")
    assert manager.storage is not None
    assert manager.current_experiment is not None  # Should create ephemeral experiment

    assert re.match(r"\d{8}_\d{6}_[a-f0-9]{8}", manager.current_experiment)
    assert manager.evaluation_progress is None


def test_session_manager_init_with_experiment(temp_storage_dir):
    """Test SessionManager initialization with experiment."""
    storage_path = f"json://{temp_storage_dir}"
    experiment_name = "test_experiment"

    manager = SessionManager(
        storage=storage_path,
        experiment_name=experiment_name,
        evaluation_name="test_eval",
    )

    assert manager.current_experiment == experiment_name
    # Check that experiment was created
    experiments = manager.storage.list_experiments()
    assert experiment_name in experiments


@patch("dotevals.sessions.get_git_commit")
def test_start_evaluation(mock_git_commit, temp_storage_dir):
    """Test starting a new evaluation."""
    mock_git_commit.return_value = "abc123"
    storage_path = f"json://{temp_storage_dir}"

    manager = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager.start_evaluation()

    assert manager.evaluation_progress is not None
    assert manager.evaluation_progress.evaluation_name == "test_eval"

    # Check that evaluation was created in storage
    loaded_eval = manager.storage.load_evaluation("test_exp", "test_eval")
    assert loaded_eval is not None
    assert loaded_eval.status == EvaluationStatus.RUNNING
    assert loaded_eval.metadata["git_commit"] == "abc123"


def test_resume_evaluation(temp_storage_dir):
    """Test resuming an existing evaluation."""
    storage_path = f"json://{temp_storage_dir}"

    # First, create an evaluation with some results
    manager1 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager1.start_evaluation()

    # Add some results
    score = Score("test_evaluator", True, [], {})
    result = Result(score, prompt="test prompt")
    record = Record(result=result, item_id=0, dataset_row={})
    manager1.add_results([record])

    # Create a new manager and resume
    manager2 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager2.start_evaluation()

    # Should detect it's resuming
    # (In real usage, this would print a message about resuming)


def test_add_results(temp_storage_dir):
    """Test adding results through SessionManager."""
    storage_path = f"json://{temp_storage_dir}"
    manager = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager.start_evaluation()

    score = Score("test_evaluator", True, [], {})
    result = Result(score, prompt="test prompt")
    record = Record(result=result, item_id=0, dataset_row={})

    manager.add_results([record])

    # Check that results were added
    results = manager.get_results()
    assert len(results) == 1
    assert results[0].item_id == 0

    # Check progress tracking
    assert manager.evaluation_progress.completed_count == 1
    assert manager.evaluation_progress.error_count == 0


def test_add_results_with_errors(temp_storage_dir):
    """Test adding results with errors."""
    storage_path = f"json://{temp_storage_dir}"
    manager = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager.start_evaluation()

    score = Score("test_evaluator", False, [], {})
    result = Result(score, prompt="test prompt")

    # Create one successful and one failed record
    record1 = Record(result=result, item_id=0, dataset_row={})
    record2 = Record(result=result, item_id=1, dataset_row={}, error="Test error")

    manager.add_results([record1, record2])

    # Check progress tracking
    assert manager.evaluation_progress.completed_count == 2
    assert manager.evaluation_progress.error_count == 1


def test_finish_evaluation(temp_storage_dir):
    """Test finishing an evaluation."""
    storage_path = f"json://{temp_storage_dir}"
    manager = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval"
    )
    manager.start_evaluation()

    # Finish successfully
    manager.finish_evaluation(success=True)

    # Check that status was updated
    loaded_eval = manager.storage.load_evaluation("test_exp", "test_eval")
    assert loaded_eval.status == EvaluationStatus.COMPLETED
    assert loaded_eval.completed_at is not None

    # Start another evaluation and finish with failure
    manager2 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="test_eval2"
    )
    manager2.start_evaluation()
    manager2.finish_evaluation(success=False)

    loaded_eval2 = manager.storage.load_evaluation("test_exp", "test_eval2")
    assert loaded_eval2.status == EvaluationStatus.FAILED


def test_finish_all_evaluations(temp_storage_dir):
    """Test finishing all active evaluations."""
    storage_path = f"json://{temp_storage_dir}"

    # Start multiple evaluations
    manager1 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="eval1"
    )
    manager2 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="eval2"
    )
    manager3 = SessionManager(
        storage=storage_path, experiment_name="test_exp", evaluation_name="eval3"
    )

    manager1.start_evaluation()
    manager2.start_evaluation()
    manager3.start_evaluation()

    # Finish all individually since we no longer have finish_all
    manager1.finish_evaluation(success=True)
    manager2.finish_evaluation(success=True)
    manager3.finish_evaluation(success=True)

    # Check all were marked as completed
    for eval_name in ["eval1", "eval2", "eval3"]:
        loaded = manager1.storage.load_evaluation("test_exp", eval_name)
        assert loaded.status == EvaluationStatus.COMPLETED


@patch("subprocess.check_output")
def test_get_git_commit_success(mock_check_output):
    """Test getting git commit when git is available."""
    mock_check_output.return_value = b"abc123def456\n"

    commit = get_git_commit()

    assert commit == "abc123de"  # First 8 characters


@patch("subprocess.check_output")
def test_get_git_commit_failure(mock_check_output):
    """Test getting git commit when git fails."""
    import subprocess

    mock_check_output.side_effect = subprocess.CalledProcessError(1, "git")

    commit = get_git_commit()

    assert commit is None


def test_session_manager_operations_with_ephemeral_experiment():
    """Test SessionManager operations with ephemeral experiment."""
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create session manager without experiment name - creates ephemeral
        session_manager = SessionManager("eval_test", storage=f"json://{temp_dir}")

        # Should have created ephemeral experiment
        assert session_manager.current_experiment is not None
        assert re.match(r"\d{8}_\d{6}_[a-f0-9]{8}", session_manager.current_experiment)

        # All operations should work with ephemeral experiment
        session_manager = SessionManager(
            storage=f"json://{temp_dir}", evaluation_name="test_eval"
        )
        session_manager.start_evaluation()

        result = Result(prompt="test")
        record = Record(result=result, item_id=0, dataset_row={})
        session_manager.add_results([record])

        results = session_manager.get_results()
        assert len(results) == 1

        session_manager.finish_evaluation(success=True)


def test_session_manager_add_results_without_progress():
    """Test adding results when evaluation_progress is None."""
    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="test_exp",
            evaluation_name="test_eval",
        )

        # Don't call start_evaluation, so evaluation_progress remains None
        session_manager.evaluation_progress = None

        # Create evaluation manually to avoid the start_evaluation logic
        evaluation = Evaluation(
            evaluation_name="test_eval",
            status=EvaluationStatus.RUNNING,
            started_at=time.time(),
        )
        session_manager.storage.create_evaluation("test_exp", evaluation)

        # Add results without evaluation_progress
        result = Result(prompt="test")
        record = Record(result=result, item_id=0, dataset_row={})

        # This should work even without evaluation_progress
        session_manager.add_results([record])

        # Verify results were added
        results = session_manager.get_results()
        assert len(results) == 1


def test_ephemeral_experiment_created_when_no_name():
    """Test that ephemeral experiment is created with timestamp when no name provided."""
    with tempfile.TemporaryDirectory() as temp_dir:
        storage_path = f"json://{temp_dir}"

        # Create session manager without experiment name
        session_manager = SessionManager("eval_test", storage=storage_path)

        # Should have created an ephemeral experiment
        assert session_manager.current_experiment is not None
        # New simplified logic: experiment name is just timestamp_uuid
        assert re.match(r"\d{8}_\d{6}_[a-f0-9]{8}", session_manager.current_experiment)

        # Verify experiment was created directly in the specified storage directory
        storage = JSONStorage(temp_dir)
        experiments = storage.list_experiments()
        assert len(experiments) == 1
        # Should contain our generated experiment name
        assert session_manager.current_experiment in experiments


def test_ephemeral_experiment_in_doteval_directory():
    """Test ephemeral experiments are created in .dotevals subdirectory."""
    with tempfile.TemporaryDirectory() as temp_dir:
        doteval_dir = Path(temp_dir) / ".dotevals"
        storage_path = f"json://{doteval_dir}"

        # Create session manager without experiment name
        session_manager = SessionManager("eval_test", storage=storage_path)

        # Should create experiment directly with timestamp_uuid
        timestamp_pattern = re.compile(r"^\d{8}_\d{6}_[a-f0-9]{8}$")
        assert timestamp_pattern.match(session_manager.current_experiment)

        # Verify directory structure
        assert doteval_dir.exists()
        exp_dir = doteval_dir / session_manager.current_experiment
        assert exp_dir.exists()


def test_ephemeral_experiment_timestamp_format():
    """Test that ephemeral experiment timestamp matches expected format."""
    with tempfile.TemporaryDirectory() as temp_dir:
        storage_path = f"json://{temp_dir}"

        # Create session manager
        before_time = datetime.now()
        session_manager = SessionManager("eval_test", storage=storage_path)
        after_time = datetime.now()

        # Extract timestamp from experiment name (before the UUID part)
        timestamp_str = session_manager.current_experiment.split("_")[:2]
        timestamp_str = "_".join(timestamp_str)

        # Parse timestamp
        timestamp = datetime.strptime(timestamp_str, "%Y%m%d_%H%M%S")

        # Verify timestamp is within expected range (ignoring microseconds)
        assert before_time.replace(microsecond=0) <= timestamp <= after_time


def test_multiple_ephemeral_experiments():
    """Test creating multiple ephemeral experiments."""
    with tempfile.TemporaryDirectory() as temp_dir:
        storage_path = f"json://{temp_dir}"

        managers = []
        for i in range(3):
            manager = SessionManager("eval_test", storage=storage_path)
            managers.append(manager)

        # All should have different experiment names
        experiment_names = [m.current_experiment for m in managers]
        assert len(set(experiment_names)) == 3

        # Verify all experiments exist in the specified storage directory
        storage = JSONStorage(temp_dir)
        ephemeral_experiments = storage.list_experiments()
        assert len(ephemeral_experiments) == 3

        # All should have timestamp_uuid format
        for exp in ephemeral_experiments:
            assert re.match(r"\d{8}_\d{6}_[a-f0-9]{8}", exp)



import json
import re
from typing import Any

from dotevals.evaluators.base import evaluator, valid_json
from dotevals.metrics import accuracy
from dotevals.models import Score
from dotevals.utils import recursive_map


def normalize_string(s):
    """Normalize string by removing special characters and converting to lowercase."""
    return re.sub(r"[ \,\./\-_\*\^]", "", str(s)).lower().replace("'", '"')


def recursive_normalize(obj: Any) -> Any:
    """
    Recursively normalize all string values in a nested structure.

    Args:
        obj: The object to normalize

    Returns:
        The normalized object with string values normalized
    """
    return recursive_map(
        obj, lambda val: normalize_string(val) if isinstance(val, str) else val
    )


def is_dict_match(actual_value, expected_value):
    """Check if the actual dictionary matches any of the expected dictionaries"""
    if isinstance(expected_value, (set | list)):
        expected_dicts = list(expected_value)
    else:
        expected_dicts = [expected_value]

    actual_norm = recursive_normalize(actual_value)
    for expected_dict in expected_dicts:
        if expected_dict == "":
            continue
        if actual_norm == recursive_normalize(expected_dict):
            return True
    if all(ed == "" for ed in expected_dicts):
        return True
    return False


def compare_param_value(param_value, expected_param_value):
    """Compare a parameter value to the expected value."""
    if isinstance(expected_param_value, dict) and isinstance(param_value, dict):
        return is_dict_match(param_value, expected_param_value)
    elif isinstance(expected_param_value, str) and isinstance(param_value, str):
        return normalize_string(param_value) == normalize_string(expected_param_value)
    elif isinstance(expected_param_value, (set | list)):
        return param_value in expected_param_value
    else:
        return param_value == expected_param_value


def is_function_call_valid(schema, model_output, expected_answer):
    """Validate that the function call matches the schema and expected answer."""
    expected_answer = list(expected_answer.values())[0]
    function_name = schema["name"]
    parameter_definitions = schema["parameters"]["properties"]
    required_parameters = set(schema["parameters"]["required"])

    if function_name not in model_output:
        return False

    model_params = model_output[function_name]

    if not required_parameters.issubset(model_params):
        return False

    try:
        json.dumps({function_name: model_params})
    except Exception:
        return False

    if not valid_json({function_name: model_params}, schema):
        return False  # pragma: no cover

    for param_name, param_value in model_params.items():
        if param_name not in parameter_definitions or param_name not in expected_answer:
            return False

        expected_param_value = expected_answer[param_name]
        if not compare_param_value(param_value, expected_param_value):
            return False

    for param_name, expected_param_value in expected_answer.items():
        if param_name not in model_params:
            if (
                isinstance(expected_param_value, (set | list))
                and "" not in expected_param_value
            ) or (
                not isinstance(expected_param_value, (set | list))
                and expected_param_value != ""
            ):
                return False

    return True


@evaluator(metrics=accuracy())
def ast_evaluation(
    result: dict[str, Any],
    expected: list[dict[str, Any]],
    schema: list[dict[str, Any]],
) -> bool:
    """Evaluate if a function call matches the expected schema and parameters."""
    if not result or not expected or not schema or len(result) != 1:
        return False

    model_output = result
    possible_answer = expected[0]
    function_description = schema[0]
    return is_function_call_valid(function_description, model_output, possible_answer)



"""Example of using the BFCL dataset for function calling evaluation.

This example requires the dotevals-datasets plugin to be installed:
    pip install dotevals-datasets

The BFCL (Berkeley Function Calling Leaderboard) dataset evaluates a model's
ability to correctly call functions based on natural language queries.

It has three variants:
- simple: Single function selection
- multiple: Select from multiple available functions
- parallel: Make multiple function calls in parallel
"""

import json

from dotevals import Result, foreach
from dotevals.evaluators import exact_match


def generate_function_call(question, available_functions):
    """Placeholder for model function calling.

    In practice, you would use your model here, e.g.:
    response = model.generate(
        question,
        tools=available_functions,
        tool_choice="auto"
    )
    """
    # For demo purposes, return empty list
    return []


@foreach.bfcl("simple")
def eval_bfcl_simple(question, schema, answer):
    """Evaluate simple function calling (single function selection).

    Args:
        question: User's natural language query
        schema: JSON string with available function definitions
        answer: JSON string with expected function call(s)
    """
    # Parse schemas and expected answer
    available_functions = json.loads(schema)
    expected_calls = json.loads(answer)

    # Generate function calls with your model
    response = generate_function_call(question, available_functions)

    # Return exact match score
    return exact_match(response, expected_calls)


@foreach.bfcl("multiple")
def eval_bfcl_multiple(question, schema, answer):
    """Evaluate multiple function selection.

    Tests if the model can select the right function from many options.
    """
    available_functions = json.loads(schema)
    expected_calls = json.loads(answer)

    response = generate_function_call(question, available_functions)

    return exact_match(response, expected_calls)


@foreach.bfcl("parallel")
def eval_bfcl_parallel(question, schema, answer):
    """Evaluate parallel function calling.

    Tests if the model can make multiple function calls for one query.
    """
    available_functions = json.loads(schema)
    expected_calls = json.loads(answer)

    response = generate_function_call(question, available_functions)

    # For parallel calls, order might not matter
    # You could implement a more sophisticated comparison here
    return exact_match(response, expected_calls)


# Advanced example with partial credit scoring
@foreach.bfcl("simple")
def eval_bfcl_with_partial_credit(question, schema, answer):
    """Evaluate with partial credit for correct function name selection.

    This gives:
    - 0.5 points for selecting the correct function
    - 0.5 points for correct parameters
    """
    available_functions = json.loads(schema)
    expected_calls = json.loads(answer)

    response = generate_function_call(question, available_functions)

    score = 0.0

    if response and expected_calls:
        # Extract function names
        response_names = {list(call.keys())[0] for call in response}
        expected_names = {list(call.keys())[0] for call in expected_calls}

        # Partial credit for correct function selection
        if response_names == expected_names:
            score = 0.5

            # Full credit if parameters also match
            if response == expected_calls:
                score = 1.0

    return Result(
        prompt=question,
        response=str(response),
        scores={
            "exact_match": score == 1.0,
            "partial_credit": score,
            "function_name_correct": score >= 0.5,
        },
    )


if __name__ == "__main__":
    # To run these evaluations:
    # pytest bfcl.py::eval_bfcl_simple --experiment bfcl_simple
    # pytest bfcl.py::eval_bfcl_multiple --experiment bfcl_multiple
    # pytest bfcl.py::eval_bfcl_parallel --experiment bfcl_parallel

    print("Run with pytest to evaluate:")
    print("  pytest bfcl.py::eval_bfcl_simple --experiment my_bfcl_eval")
    print("\nView results with:")
    print("  dotevals show my_bfcl_eval")



"""Consolidated evaluator tests - more concise with parameterized testing."""

import json
import tempfile
from typing import Any

import pytest

from dotevals.evaluators import evaluator, exact_match, numeric_match, valid_json
from dotevals.evaluators.ast_evaluation import ast_evaluation
from dotevals.evaluators.registry import EvaluatorRegistry
from dotevals.metrics import accuracy
from dotevals.models import Score


class TestExactMatch:
    """Tests for exact_match evaluator."""

    @pytest.mark.parametrize(
        "answer,expected,result",
        [
            ("1", "1", True),
            ("1", "2", False),
            ("hello", "hello", True),
            ("Hello", "hello", False),
            (1, 1, True),
            (1, 2, False),
            (None, None, True),
            (None, "", False),
        ],
    )
    def test_exact_match_basic(self, answer, expected, result):
        """Test basic exact match comparisons."""
        score = exact_match(answer, expected)
        assert isinstance(score, Score)
        assert score.value is result
        assert score.name == "exact_match"
        assert len(score.metrics) == 1
        assert score.metrics[0].__name__ == "accuracy"

    @pytest.mark.parametrize("custom_name", ["custom_comparison", "my_evaluator", None])
    def test_exact_match_custom_name(self, custom_name):
        """Test custom naming for exact_match."""
        score = exact_match("test", "test", name=custom_name)
        expected_name = custom_name if custom_name else "exact_match"
        assert score.name == expected_name


class TestNumericMatch:
    """Tests for numeric_match evaluator."""

    @pytest.mark.parametrize(
        "answer,expected,should_match",
        [
            # Basic integers
            (42, 42, True),
            (42, 43, False),
            ("42", "42", True),
            ("42", 42, True),
            (42, "42", True),
            # Floats
            (3.14, 3.14, True),
            ("3.14", 3.14, True),
            (3.14, "3.14", True),
            (3.14, 3.15, False),
            # Numbers with commas
            ("1,234", "1234", True),
            ("1,234,567", "1234567", True),
            ("1,234.56", "1234.56", True),
            # Numbers with spaces
            ("1 234", "1234", True),
            ("1 234 567", "1234567", True),
            ("1 234.56", "1234.56", True),
            # Scientific notation
            ("1.23e3", "1230", True),
            ("1.23E3", "1230", True),
            ("1.23e-3", "0.00123", True),
            # Special cases
            ("0", 0, True),
            ("-42", -42, True),
            ("", "", False),
            ("abc", "123", False),
            ("12.00", "12", True),
            ("1.0", "1", True),
            # Leading/trailing spaces
            ("  42  ", "42", True),
            ("\t100\n", "100", True),
            # None values (edge cases)
            (None, "42", False),
            ("42", None, False),
            (None, None, False),
            # Space after minus sign (invalid number format)
            ("- 42", "42", False),
            ("- 42", "- 42", False),
        ],
    )
    def test_numeric_match_formats(self, answer, expected, should_match):
        """Test various numeric formats and comparisons."""
        score = numeric_match(answer, expected)
        assert score.value is should_match
        assert score.name == "numeric_match"

    def test_numeric_match_custom_name(self):
        """Test custom naming for numeric_match."""
        score = numeric_match(42, 42, name="number_check")
        assert score.name == "number_check"


class TestValidJson:
    """Tests for valid_json evaluator."""

    @pytest.mark.parametrize(
        "text,is_valid",
        [
            # Valid JSON
            ('{"key": "value"}', True),
            ("[]", True),
            ("null", True),
            ("true", True),
            ("false", True),
            ("42", True),
            ('"string"', True),
            ('{"nested": {"key": "value"}}', True),
            ("[1, 2, 3]", True),
            # Invalid JSON
            ("", False),
            ("{", False),
            ('{"key": }', False),
            ("{'key': 'value'}", False),  # Single quotes
            ("undefined", False),
            ('{key: "value"}', False),  # Unquoted key
            ('{"key": "value",}', False),  # Trailing comma
            # Special cases
            ('  {"key": "value"}  ', True),  # With whitespace
            ('{"unicode": "🎉"}', True),  # Unicode
        ],
    )
    def test_valid_json_basic(self, text, is_valid):
        """Test JSON validation."""
        score = valid_json(text)
        assert score.value is is_valid
        assert score.name == "valid_json"

    def test_valid_json_custom_name(self):
        """Test custom naming for valid_json."""
        score = valid_json('{"valid": true}', name="json_check")
        assert score.name == "json_check"

    def test_valid_json_with_schema(self):
        """Test valid_json with JSON schema validation."""
        schema = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
            "required": ["name"],
        }

        # Valid JSON matching schema
        score = valid_json('{"name": "Alice", "age": 30}', schema=schema)
        assert score.value is True

        # Valid JSON not matching schema (missing required field)
        score = valid_json('{"age": 30}', schema=schema)
        assert score.value is False

        # Invalid schema
        score = valid_json('{"name": "Alice"}', schema={"type": "invalid_type"})
        assert score.value is False

    def test_valid_json_none_input(self):
        """Test valid_json with None input."""
        # Test None value
        score = valid_json(None)
        assert score.value is False


class TestAstEvaluation:
    """Tests for ast_evaluation evaluator with function call schema matching."""

    def test_ast_evaluation_exact_match(self):
        """Test exact function call matching."""
        result = {"calculate": {"x": 5, "y": 3}}
        expected = [{"calculate": {"x": 5, "y": 3}}]
        schema = [
            {
                "name": "calculate",
                "parameters": {
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is True
        assert score.name == "ast_evaluation"

    def test_ast_evaluation_wrong_function(self):
        """Test mismatch when function name is wrong."""
        result = {"wrong": {"x": 5, "y": 3}}
        expected = [{"calculate": {"x": 5, "y": 3}}]
        schema = [
            {
                "name": "calculate",
                "parameters": {
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is False

    def test_ast_evaluation_missing_parameter(self):
        """Test mismatch when required parameter is missing."""
        result = {"calculate": {"x": 5}}
        expected = [{"calculate": {"x": 5, "y": 3}}]
        schema = [
            {
                "name": "calculate",
                "parameters": {
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is False

    def test_ast_evaluation_string_normalization(self):
        """Test string normalization in comparisons."""
        result = {"greet": {"name": "Alice", "msg": "HELLO"}}
        expected = [{"greet": {"name": "alice", "msg": "hello"}}]
        schema = [
            {
                "name": "greet",
                "parameters": {
                    "properties": {
                        "name": {"type": "string"},
                        "msg": {"type": "string"},
                    },
                    "required": ["name", "msg"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is True

    def test_ast_evaluation_set_alternatives(self):
        """Test matching against set of alternatives."""
        result = {"choose": {"color": "red"}}
        expected = [{"choose": {"color": {"red", "blue", "green"}}}]
        schema = [
            {
                "name": "choose",
                "parameters": {
                    "properties": {"color": {"type": "string"}},
                    "required": ["color"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is True

        # Test non-matching alternative
        result = {"choose": {"color": "yellow"}}
        score = ast_evaluation(result, expected, schema)
        assert score.value is False

    def test_ast_evaluation_optional_param(self):
        """Test optional parameters."""
        result = {"greet": {"name": "Alice"}}
        expected = [{"greet": {"name": "Alice", "greeting": ""}}]
        schema = [
            {
                "name": "greet",
                "parameters": {
                    "properties": {
                        "name": {"type": "string"},
                        "greeting": {"type": "string"},
                    },
                    "required": ["name"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is True

    def test_ast_evaluation_helper_functions(self):
        """Test helper functions for ast_evaluation."""
        from dotevals.evaluators.ast_evaluation import (
            compare_param_value,
            is_dict_match,
            recursive_normalize,
        )

        # Test recursive_normalize
        assert recursive_normalize("Hello") == "hello"
        assert recursive_normalize(["A", "B"]) == ["a", "b"]
        assert recursive_normalize({"key": "VALUE"}) == {"key": "value"}
        assert recursive_normalize(42) == 42

        # Test is_dict_match
        assert is_dict_match({"a": 1}, {"a": 1}) is True
        assert is_dict_match({"a": 1}, {"a": 2}) is False
        assert is_dict_match({"a": 1, "b": 2}, {"a": 1}) is False

        # Test compare_param_value
        assert compare_param_value("test", "test") is True
        assert compare_param_value("TEST", "test") is True  # normalized
        assert compare_param_value("red", {"red", "blue"}) is True
        assert compare_param_value("green", {"red", "blue"}) is False

        # Test edge cases for is_dict_match
        assert is_dict_match({"a": 1}, [{"a": 1}]) is True  # list of expected
        assert is_dict_match({"a": 1}, "") is True  # empty string expected returns True
        assert is_dict_match({"a": 1}, [""]) is True  # all empty strings

        # Test edge case for compare_param_value
        assert (
            compare_param_value({"key": "val"}, {"key": "val"}) is True
        )  # dict comparison

    def test_ast_evaluation_json_serialization_error(self):
        """Test when model output contains non-serializable objects ."""

        # Create a non-serializable object
        class NonSerializable:
            pass

        result = {"func": {"param": NonSerializable()}}
        expected = [{"func": {"param": "value"}}]
        schema = [
            {
                "name": "func",
                "parameters": {
                    "properties": {"param": {"type": "string"}},
                    "required": ["param"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is False  # Should fail due to JSON serialization error

    def test_ast_evaluation_missing_expected_param(self):
        """Test when expected param is missing in model output."""
        result = {"greet": {"name": "Alice"}}
        expected = [{"greet": {"name": "Alice", "age": 30}}]  # age is required
        schema = [
            {
                "name": "greet",
                "parameters": {
                    "properties": {
                        "name": {"type": "string"},
                        "age": {"type": "number"},
                    },
                    "required": ["name", "age"],
                },
            }
        ]
        score = ast_evaluation(result, expected, schema)
        assert score.value is False  # Should fail due to missing age param


class TestEvaluatorDecorator:
    """Tests for the @evaluator decorator."""

    def test_evaluator_basic(self):
        """Test basic evaluator decorator usage."""

        @evaluator(metrics=accuracy())
        def custom_eval(answer, expected):
            return answer == expected

        score = custom_eval("test", "test")
        assert isinstance(score, Score)
        assert score.value is True
        assert score.name == "custom_eval"
        assert len(score.metrics) == 1

    def test_evaluator_multiple_metrics(self):
        """Test evaluator with multiple metrics."""
        from dotevals.metrics import metric

        @metric
        def always_one():
            def metric_fn(scores):
                return 1.0

            return metric_fn

        @evaluator(metrics=[accuracy(), always_one()])
        def multi_metric_eval(a, b):
            return a == b

        score = multi_metric_eval(1, 1)
        assert len(score.metrics) == 2
        assert score.metrics[0].__name__ == "accuracy"
        assert score.metrics[1].__name__ == "always_one"

    def test_evaluator_with_metadata(self):
        """Test that evaluator captures metadata."""

        @evaluator(metrics=accuracy())
        def eval_with_metadata(answer, expected, threshold=0.5):
            return answer >= threshold

        score = eval_with_metadata(0.7, None, threshold=0.6)
        assert score.metadata == {"answer": 0.7, "expected": None, "threshold": 0.6}

    def test_evaluator_custom_name_override(self):
        """Test that custom name overrides function name."""

        @evaluator(metrics=accuracy())
        def my_evaluator(a, b):
            return a == b

        score = my_evaluator(1, 1, name="custom_name")
        assert score.name == "custom_name"


class TestEvaluatorRegistry:
    """Tests for the evaluator registry."""

    def test_registry_has_builtin_evaluators(self):
        """Test registry auto-discovers built-in evaluators."""
        registry = EvaluatorRegistry()
        evaluators = registry.list_evaluators()

        # Should have built-in evaluators
        assert "exact_match" in evaluators
        assert "numeric_match" in evaluators
        assert "valid_json" in evaluators
        assert "ast_evaluation" in evaluators

    def test_registry_registration(self):
        """Test registering custom evaluators."""
        registry = EvaluatorRegistry()
        initial_count = len(registry.list_evaluators())

        def dummy_evaluator():
            return Score("dummy", True, [])

        registry.register("dummy", dummy_evaluator)
        assert "dummy" in registry.list_evaluators()
        assert registry.get("dummy") == dummy_evaluator
        assert len(registry.list_evaluators()) == initial_count + 1

    def test_registry_get_builtin(self):
        """Test getting built-in evaluators."""
        registry = EvaluatorRegistry()

        # Should be able to get built-in evaluators
        assert registry.get("exact_match") is not None
        assert registry.get("numeric_match") is not None
        assert registry.get("valid_json") is not None

    def test_registry_get_nonexistent(self):
        """Test getting non-existent evaluator returns None."""
        registry = EvaluatorRegistry()
        assert registry.get("nonexistent") is None

    def test_registry_duplicate_registration(self):
        """Test that duplicate registration overwrites."""
        registry = EvaluatorRegistry()

        def evaluator1():
            return Score("eval1", True, [])

        def evaluator2():
            return Score("eval2", False, [])

        registry.register("test_dup", evaluator1)
        assert registry.get("test_dup") == evaluator1

        registry.register("test_dup", evaluator2)
        assert registry.get("test_dup") == evaluator2

    def test_registry_discover_plugins(self):
        """Test plugin discovery mechanism."""
        registry = EvaluatorRegistry()
        initial_evaluators = registry.list_evaluators()

        # Registry auto-discovers on init, so should have evaluators
        assert len(initial_evaluators) > 0
        assert "exact_match" in initial_evaluators
        assert "numeric_match" in initial_evaluators

    def test_registry_methods_exist(self):
        """Test that registry has expected methods."""
        registry = EvaluatorRegistry()
        # Check core methods exist
        assert hasattr(registry, "register")
        assert hasattr(registry, "get")
        assert hasattr(registry, "list_evaluators")
        # Registry should have evaluators already discovered
        assert len(registry.list_evaluators()) > 0


def test_evaluator_registry_plugin_loading():
    """Test evaluator registry plugin loading."""
    from unittest.mock import MagicMock, patch

    from dotevals.evaluators.registry import EvaluatorRegistry

    registry = EvaluatorRegistry()

    # Mock entry points
    mock_entry_point = MagicMock()
    mock_entry_point.name = "plugin_evaluator"
    mock_entry_point.load.return_value = lambda x: x  # Simple evaluator

    with patch("importlib.metadata.entry_points") as mock_entry_points:
        mock_entry_points.return_value = {"dotevals.evaluators": [mock_entry_point]}

        # Load plugins
        registry.load_plugins()

        # Check plugin was loaded
        assert "plugin_evaluator" in registry.list_evaluators()
        assert registry.get("plugin_evaluator") is not None


def test_evaluator_registry_plugin_error():
    """Test evaluator registry handles plugin errors."""
    from unittest.mock import MagicMock, patch

    from dotevals.evaluators.registry import EvaluatorRegistry

    registry = EvaluatorRegistry()

    # Mock entry points that fail to load
    mock_entry_point = MagicMock()
    mock_entry_point.name = "bad_plugin"
    mock_entry_point.load.side_effect = ImportError("Plugin failed")

    with patch("importlib.metadata.entry_points") as mock_entry_points:
        mock_entry_points.return_value = {"dotevals.evaluators": [mock_entry_point]}

        # Should handle error gracefully
        registry.load_plugins()

        # Bad plugin should not be loaded
        assert "bad_plugin" not in registry.list_evaluators()


def test_evaluator_module_imports():
    """Test dynamic imports from evaluators module."""
    # Test __getattr__ for successful import
    from dotevals import evaluators

    em = evaluators.exact_match
    assert em is not None

    # Test __dir__ listing
    available = dir(evaluators)
    assert "exact_match" in available
    assert "numeric_match" in available

    # Test AttributeError for non-existent evaluator
    import pytest

    with pytest.raises(AttributeError, match="No evaluator named 'nonexistent'"):
        _ = evaluators.nonexistent


def test_additional_evaluator_coverage():
    """Test additional evaluator functionality for coverage."""
    # Test more numeric_match edge cases
    score = numeric_match(float("inf"), float("inf"))
    assert score.value is True

    score = numeric_match(float("nan"), float("nan"))
    assert score.value is False

    # Test exact_match with custom types
    class CustomObj:
        def __eq__(self, other):
            return True

    obj = CustomObj()
    score = exact_match(obj, "anything")
    assert score.value is True


def test_ast_evaluation_edge_cases():
    """Test ast_evaluation edge cases for better coverage."""
    from dotevals.evaluators.ast_evaluation import (
        ast_evaluation,
        compare_param_value,
        recursive_normalize,
    )

    # Test list normalization in recursive_normalize
    assert recursive_normalize({"items": ["A", "B"]}) == {"items": ["a", "b"]}
    assert recursive_normalize({"nested": {"key": "VALUE"}}) == {
        "nested": {"key": "value"}
    }
    assert recursive_normalize(42) == 42  # Numbers unchanged
    assert recursive_normalize(None) is None  # None unchanged

    # Test compare_param_value with list expected values
    assert compare_param_value({"a": 1}, [{"a": 1}, {"b": 2}]) is True
    assert compare_param_value({"c": 3}, [{"a": 1}, {"b": 2}]) is False

    # Test with empty result (no function call)
    result = {}
    expected = [{"func": {}}]
    schema = [{"name": "func", "parameters": {}}]
    score = ast_evaluation(result, expected, schema)
    assert score.value is False

    # Test with extra parameters (should fail)
    result = {"calculate": {"x": 5, "y": 3, "z": 1}}
    expected = [{"calculate": {"x": 5, "y": 3}}]
    schema = [
        {
            "name": "calculate",
            "parameters": {
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
            },
        }
    ]
    score = ast_evaluation(result, expected, schema)
    assert score.value is False


def test_registry_plugin_loading():
    """Test registry plugin loading mechanisms."""
    from unittest.mock import Mock, patch

    from dotevals.evaluators.registry import EvaluatorRegistry
    from dotevals.models import Score

    # Test with mock entry points
    mock_entry_point = Mock()
    mock_entry_point.name = "test_evaluator"

    def test_eval_func():
        return Score("test", True, [])

    mock_entry_point.load.return_value = test_eval_func

    with patch("importlib.metadata.entry_points") as mock_eps:
        # Mock entry_points to return our test evaluator
        mock_group = Mock()
        mock_group.select.return_value = [mock_entry_point]
        mock_eps.return_value = mock_group

        registry = EvaluatorRegistry()
        # Since registry auto-discovers, test_evaluator should be there
        # Or manually discover to ensure it's loaded
        evaluators = registry.list_evaluators()
        # Just verify the registry works with mocked entry points
        assert isinstance(evaluators, list)


def test_evaluator_basic_integration():
    """Basic integration test for key evaluators."""
    # Test exact_match
    score = exact_match("hello", "hello")
    assert score.value is True
    assert score.name == "exact_match"

    # Test numeric_match
    score = numeric_match("1,234.56", 1234.56)
    assert score.value is True
    assert score.name == "numeric_match"

    # Test valid_json
    score = valid_json('{"valid": true}')
    assert score.value is True
    assert score.name == "valid_json"

    # Test ast_evaluation with proper schema
    from dotevals.evaluators.ast_evaluation import ast_evaluation as ast_eval_func

    result = {"calculate": {"x": 5, "y": 3}}
    expected = [{"calculate": {"x": 5, "y": 3}}]
    schema = [
        {
            "name": "calculate",
            "parameters": {
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
            },
        }
    ]
    score = ast_eval_func(result, expected, schema)
    assert score.value is True
    assert score.name == "ast_evaluation"



"""Concurrency strategies for dotevals."""

from typing import Union

from .adaptive import Adaptive
from .async_sequential import AsyncSequential
from .batch import Batch
from .sequential import Sequential
from .sliding_window import SlidingWindow

# Backward compatibility aliases (will be deprecated in future versions)
SyncConcurrencyStrategy = Batch | Sequential
AsyncConcurrencyStrategy = SlidingWindow | Adaptive | AsyncSequential

__all__ = [
    "Adaptive",
    "AsyncSequential",
    "Batch",
    "Sequential",
    "SlidingWindow",
    # Type unions
    "SyncConcurrencyStrategy",
    "AsyncConcurrencyStrategy",
]



"""Tests for error handling in metric calculations."""

import asyncio
import tempfile

from dotevals.core import foreach
from dotevals.evaluators import exact_match
from dotevals.models import EvaluationSummary, Record, Result
from dotevals.sessions import SessionManager


def test_errors_included_in_metrics():
    """Test that errors are counted as False in metrics."""
    dataset = [
        ("test1", "answer1"),
        ("test2", "answer2"),
        ("test3", "answer3"),
        ("test4", "answer4"),
    ]

    @foreach("question,answer", dataset)
    def eval_with_errors(question, answer):
        if question == "test2":
            raise ValueError("Simulated error")
        if question == "test4":
            raise RuntimeError("Another error")
        return Result(exact_match(answer, "answer1"), prompt=question)

    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="error_metrics_test",
            evaluation_name="test_eval",
        )
        coro = eval_with_errors(session_manager=session_manager, samples=None)
        result = asyncio.run(coro)

        # We should have 4 results (2 successful, 2 errors)
        assert len(result.results) == 4

        # Check that errors are included in metric calculation
        # Only test1 matches "answer1", test3 doesn't match, test2 and test4 are errors
        # So accuracy should be 1/4 = 0.25
        assert result.summary["exact_match"]["accuracy"] == 0.25

        # Verify error results exist
        error_results = [r for r in result.results if r.error is not None]
        assert len(error_results) == 2


def test_all_errors_zero_accuracy():
    """Test that all errors result in zero accuracy."""
    dataset = [("test1", "answer1"), ("test2", "answer2")]

    @foreach("question,answer", dataset)
    def eval_all_errors(question, answer):
        raise Exception("Always fail")

    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="all_errors_test",
            evaluation_name="test_eval",
        )
        coro = eval_all_errors(session_manager=session_manager, samples=None)
        result = asyncio.run(coro)

        # All items should have errors
        assert len(result.results) == 2
        assert all(r.error is not None for r in result.results)
        # When all results are errors, there's no score structure to infer
        # So the summary will be empty
        assert result.summary == {}


def test_no_errors_normal_accuracy():
    """Test that no errors results in normal accuracy calculation."""
    dataset = [
        ("test1", "answer1"),
        ("test2", "answer2"),
        ("test3", "answer3"),
    ]

    @foreach("question,answer", dataset)
    def eval_no_errors(question, answer):
        # Match only answer2
        return Result(exact_match(answer, "answer2"), prompt=question)

    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="no_errors_test",
            evaluation_name="test_eval",
        )
        coro = eval_no_errors(session_manager=session_manager, samples=None)
        result = asyncio.run(coro)

        # No errors, normal calculation
        assert len(result.results) == 3
        assert all(r.error is None for r in result.results)
        # Only test2 matches "answer2"
        assert result.summary["exact_match"]["accuracy"] == 1 / 3


def test_summary_includes_both_metrics():
    """Test that summary includes metrics with and without errors."""
    # Create mock results
    results = [
        Record(
            result=Result(exact_match("a", "a"), prompt="test1"),  # True
            item_id=0,
            dataset_row={"q": "test1"},
            error=None,
        ),
        Record(
            result=Result(exact_match("b", "a"), prompt="test2"),  # False
            item_id=1,
            dataset_row={"q": "test2"},
            error=None,
        ),
        Record(
            result=Result(prompt=""),  # Error - will be changed to False
            item_id=2,
            dataset_row={"q": "test3"},
            error="ConnectionError",
        ),
        Record(
            result=Result(exact_match("a", "a"), prompt="test4"),  # True
            item_id=3,
            dataset_row={"q": "test4"},
            error=None,
        ),
    ]

    # Create summary
    summary = EvaluationSummary(results)
    summary_dict = summary.compute_summary()

    # With errors included as False:
    # 2 True results out of 4 total = 0.5
    assert "exact_match" in summary_dict
    assert summary_dict["exact_match"]["accuracy"] == 0.5


def test_mixed_errors_and_successes():
    """Test evaluation with mix of errors and successful evaluations."""
    dataset = [
        ("q1", "a1"),
        ("q2", "a2"),
        ("q3", "a3"),
        ("q4", "a4"),
        ("q5", "a5"),
    ]

    error_questions = {"q2", "q4"}

    @foreach("question,answer", dataset)
    def eval_mixed(question, answer):
        if question in error_questions:
            raise Exception(f"Error on {question}")
        # Only q1 and q5 match their answers
        return Result(exact_match(question, question), prompt=question)

    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="mixed_test",
            evaluation_name="test_eval",
        )
        coro = eval_mixed(session_manager=session_manager, samples=None)
        result = asyncio.run(coro)

        # All 5 items should be in results
        assert len(result.results) == 5

        # Count errors
        error_count = sum(1 for r in result.results if r.error is not None)
        assert error_count == 2

        # With errors counted as False:
        # q1: True, q2: False (error), q3: True, q4: False (error), q5: True
        # Accuracy should be 3/5 = 0.6
        assert result.summary["exact_match"]["accuracy"] == 0.6


def test_error_preserves_dataset_row():
    """Test that error results preserve dataset row information."""
    dataset = [("question_text", "answer_text", "extra_info")]

    @foreach("q,a,info", dataset)
    def eval_preserve_row(q, a, info):
        raise ValueError("Test error")

    with tempfile.TemporaryDirectory() as temp_dir:
        session_manager = SessionManager(
            storage=f"json://{temp_dir}",
            experiment_name="preserve_row_test",
            evaluation_name="test_eval",
        )
        coro = eval_preserve_row(session_manager=session_manager, samples=None)
        result = asyncio.run(coro)

        assert len(result.results) == 1
        error_result = result.results[0]

        # Dataset row should be preserved
        assert error_result.dataset_row == {
            "q": "question_text",
            "a": "answer_text",
            "info": "extra_info",
        }
        assert error_result.error == "ValueError: Test error"
